"""
scrapli_netconf_call
####################

Dependencies: `Scrapli Netconf <https://pypi.org/project/scrapli-netconf/>`_ need to be installed

Scrapli Netconf is a library to interact with devices using NETCONF. ``scrapli_netconf_call`` task 
plugin is a wrapper around Scrapli Netconf connection object.

NETCONF protocol has a specific set of RPC calls available for use, rather
than coding separate task for each of them, ``scrapli_netconf_call`` made to execute
any arbitrary method supported by Scrapli Netconf connection object plus a set 
of additional helper methods for extended functionality.

Sample code to run ``scrapli_netconf_call`` task::

    from nornir import InitNornir
    from nornir_salt import scrapli_netconf_call

    nr = InitNornir(config_file="config.yaml")
    
    output = nr.run(
        task=scrapli_netconf_call,
        call="get_config",
        source="running"
    )

API Reference
=============

.. autofunction:: nornir_salt.plugins.tasks.scrapli_netconf_call.scrapli_netconf_call

Additional Call Methods Reference
=================================

dir
---
.. autofunction:: nornir_salt.plugins.tasks.scrapli_netconf_call._call_dir

help
----
.. autofunction:: nornir_salt.plugins.tasks.scrapli_netconf_call._call_help

server_capabilities
-------------------

.. autofunction:: nornir_salt.plugins.tasks.scrapli_netconf_call._call_server_capabilities

connected
---------

.. autofunction:: nornir_salt.plugins.tasks.scrapli_netconf_call._call_connected

locked
------

.. autofunction:: nornir_salt.plugins.tasks.scrapli_netconf_call._call_locked
"""
import logging
import traceback
from nornir.core.task import Result, Task

log = logging.getLogger(__name__)

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "scrapli_netconf"


def _call_dir(conn, *args, **kwargs):
    """Helper function to return a list of supported tasks"""
    methods = [m for m in dir(conn) if not m.startswith("_")] + [
        "dir",
        "connected",
        "locked",
    ]
    return sorted(methods), False


def _call_help(conn, *args, **kwargs):
    """
    Helper function to return docstring for requested method

    :param method_name: (str) name of method or function to return docstring for
    """
    method_name = kwargs["method_name"]
    if "_call_{}".format(method_name) in globals():
        function_obj = globals()["_call_{}".format(method_name)]
    else:
        function_obj = getattr(conn, method_name)
    h = function_obj.__doc__ if hasattr(function_obj, "__doc__") else ""
    return h, False


def _call_connected(conn, *args, **kwargs):
    """Helper function to return connection status"""
    return conn.isalive(), False


def _call_locked(conn, *args, **kwargs):
    """
    Helper function to run this edit-config flow:

    1. Lock target configuration/datastore
    2. Discard previous changes if any
    3. Run edit config
    4. Validate new confiugration if server supports it
    5. Commit configuration
    6. Unlock target configuration/datastore

    If any of steps 3, 4, 5 fails, all changes discarded

    :param target: (str) name of datastore to edit configuration for
    :param config: (str) configuration to apply
    :returns result: (list) list of steps performed with details
    :returns failed: (bool) status indicator if change failed
    """
    failed = False
    result = []
    try:
        # lock target config/datastore
        r = conn.lock(target=kwargs["target"])
        result.append({"lock": r.result})
        # discard previous changes if any
        r = conn.discard()
        result.append({"discard_changes": r.result})
        # apply configuration
        r = conn.edit_config(config=kwargs["config"], target=kwargs["target"])
        result.append({"edit_config": r.result})
        # validate configuration
        try:
            r = conn.validate(source=kwargs["target"])
            result.append({"validate": r.result})
        except Exception as e:
            result.append({"validate": str(e)})
            pass
        # commit configuration
        r = conn.commit()
        result.append({"commit": r.result})
        # unlock target config/datastore
        r = conn.unlock(target=kwargs["target"])
        result.append({"unlock": r.result})
    except:
        # form error message
        tb = traceback.format_exc()
        log.error(
            "nornir_salt:scrapli_netconf_call locked edit_config call error: {}".format(
                tb
            )
        )
        result.append({"error": tb})
        failed = True
        # discard changes on failure
        r = conn.discard()
        result.append({"discard_changes": r.result})
        # unlock target config/datastore
        r = conn.unlock(target=kwargs["target"])
        result.append({"unlock": r.result})

    return result, failed


def _call_server_capabilities(conn, *args, **kwargs):
    """Helper function to return NETCONF server capabilities"""
    return conn.server_capabilities, False


def scrapli_netconf_call(task: Task, call: str, *args, **kwargs) -> Result:
    """
    Dispatcher function to call one of the supported scrapli_netconf methods
    or one of helper functions.

    :param call: (str) Scrapli Netconf connection object method to call
    :param arg: (list) any ``*args`` to use with call method
    :param kwargs: (dict) any ``**kwargs`` to use with call method
    :return: result of scrapli-netconf connection method call
    """
    # initiate local parameteres
    result = None
    failed = False
    task.name = call

    # get rendered data if any
    if "__task__" in task.host.data:
        kwargs.update(task.host.data["__task__"])

    log.debug(
        "nornir_salt:scrapli_netconf_call calling '{}' with args: '{}'; kwargs: '{}'".format(
            call, args, kwargs
        )
    )

    # get scrapli-netconf connection object
    conn = task.host.get_connection("scrapli_netconf", task.nornir.config)

    # check if need to call one of helper function
    if "_call_{}".format(call) in globals():
        result, failed = globals()["_call_{}".format(call)](conn, *args, **kwargs)
    # call conn object method otherwise
    else:
        result = getattr(conn, call)(*args, **kwargs)
        failed = result.failed if hasattr(result, "failed") else failed

    # format results
    if hasattr(result, "result"):
        result = result.result
    elif isinstance(result, (list, dict, bool)):
        pass
    else:
        result = str(result)

    return Result(host=task.host, result=result, failed=failed)



"""
sleep
#####

Function to sleep for given amount of time on a per-host basis.

Code to invoke ``sleep`` task::

    from nornir_salt import sleep

    output = nr.run(task=sleep, sleep_for=10)

Returns Nornir results object with task results string "Slept for {sleep_for}s".

API Reference
=============

.. autofunction:: nornir_salt.plugins.tasks.sleep.sleep
"""
import time
from nornir.core.task import Result


def sleep(task, sleep_for=5):
    """
    Task to sleep for given amount of time.

    :param sleep_for: (int) time in seconds to sleep for
    """
    task.name = "sleep"
    time.sleep(int(sleep_for))

    return Result(host=task.host, result="Slept for {}s".format(sleep_for))



"""
SaltEventProcessor Plugin
#########################

Processor plugin to emit events on task execution progress,
used by Nornir Proxy Runner modules to track tasks flow.

SaltEventProcessor does not work outside of SaltStack environment.

SaltEventProcessor reference
============================

.. autofunction:: nornir_salt.plugins.processors.SaltEventProcessor.SaltEventProcessor
"""
import logging
import traceback
import time

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task

log = logging.getLogger(__name__)

try:
    # starting with salt 3003 need to use loader_context to reconstruct
    # __salt__ dunder within treads:
    # details: https://github.com/saltstack/salt/issues/59962
    try:
        from salt.loader_context import loader_context

    except ImportError:
        # after salt 3004 api was updated - https://github.com/saltstack/salt/pull/60595
        from salt.loader.context import loader_context

    HAS_LOADER_CONTEXT = True
except ImportError:
    HAS_LOADER_CONTEXT = False


class SaltEventProcessor:
    """
    SaltEventProcessor can emit event on SaltStack Event bus about task execution progress.

    :param __salt__: (obj) __salt__ dunder object
    :param loader: (obj) salt loader, required to use __salt__ dunder within threads
        for SaltStack version above 3003.
    :param loader_context: (obj) salt loader context
    :param proxy_id: (str) Proxy Minion ID to form event tags
    :param tftr: (str) timestamp formatter string, default is "%d-%b-%Y %H:%M:%S"
    :param identity: (dict) task identity dictionary of uuid4, jid, function_name keys
    """

    def __init__(self, __salt__, loader, proxy_id, identity, tftr="%d-%b-%Y %H:%M:%S"):
        self.__salt__ = __salt__
        self.loader = loader
        self.proxy_id = proxy_id
        self.tftr = tftr
        self.jid = identity["jid"]
        self.function = identity["function_name"]

    def _emit_event(self, tag, data):
        """
        Helper function to emit event on SaltStack Event BUS.

        :param tag: (str) event tag string
        :param data: (any) event data content
        """
        if HAS_LOADER_CONTEXT and self.loader is not None:
            with loader_context(self.loader):
                self.__salt__["event.send"](tag=tag, data=data)
        else:
            self.__salt__["event.send"](tag=tag, data=data)

    def _timestamp(self):
        """
        Helper function to produce event data timestamp.
        """
        return time.strftime(self.tftr)

    def task_started(self, task: Task) -> None:
        tag = "nornir-proxy/{jid}/{proxy_id}/task/started/{task_name}".format(
            proxy_id=self.proxy_id, task_name=task.name, jid=self.jid
        )
        data = {
            "timestamp": self._timestamp(),
            "task_name": task.name,
            "jid": self.jid,
            "proxy_id": self.proxy_id,
            "task_event": "started",
            "task_type": "task",
            "hosts": list(task.nornir.inventory.hosts.keys()),
            "status": "RUNNING",
            "function": self.function,
        }
        self._emit_event(tag, data)

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        tag = "nornir-proxy/{jid}/{proxy_id}/task/completed/{task_name}".format(
            proxy_id=self.proxy_id, task_name=task.name, jid=self.jid
        )
        data = {
            "timestamp": self._timestamp(),
            "task_name": task.name,
            "jid": self.jid,
            "proxy_id": self.proxy_id,
            "task_event": "completed",
            "task_type": "task",
            "hosts": list(task.nornir.inventory.hosts.keys()),
            "status": "FAILED" if task.results.failed else "PASSED",
            "function": self.function,
        }
        self._emit_event(tag, data)

    def task_instance_started(self, task: Task, host: Host) -> None:
        tag = "nornir-proxy/{jid}/{proxy_id}/{host}/task/started/{task_name}".format(
            proxy_id=self.proxy_id,
            host=host.name,
            task_name=task.name,
            jid=self.jid,
        )
        data = {
            "timestamp": self._timestamp(),
            "task_name": task.name,
            "jid": self.jid,
            "host": host.name,
            "proxy_id": self.proxy_id,
            "task_event": "started",
            "task_type": "task_instance",
            "status": "RUNNING",
            "function": self.function,
        }
        self._emit_event(tag, data)

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        tag = "nornir-proxy/{jid}/{proxy_id}/{host}/task/completed/{task_name}".format(
            proxy_id=self.proxy_id,
            host=host.name,
            task_name=task.name,
            jid=self.jid,
        )
        data = {
            "timestamp": self._timestamp(),
            "task_name": task.name,
            "jid": self.jid,
            "proxy_id": self.proxy_id,
            "host": host.name,
            "task_event": "completed",
            "task_type": "task_instance",
            "status": "FAILED" if task.results.failed else "PASSED",
            "function": self.function,
        }
        self._emit_event(tag, data)

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        tag = "nornir-proxy/{jid}/{proxy_id}/{host}/subtask/started/{task_name}".format(
            proxy_id=self.proxy_id,
            host=host.name,
            task_name=task.name,
            jid=self.jid,
        )
        data = {
            "timestamp": self._timestamp(),
            "task_name": task.name,
            "jid": self.jid,
            "proxy_id": self.proxy_id,
            "host": host.name,
            "task_event": "started",
            "task_type": "subtask",
            "status": "RUNNING",
            "function": self.function,
        }
        self._emit_event(tag, data)

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        tag = (
            "nornir-proxy/{jid}/{proxy_id}/{host}/subtask/completed/{task_name}".format(
                proxy_id=self.proxy_id,
                host=host.name,
                task_name=task.name,
                jid=self.jid,
            )
        )
        data = {
            "timestamp": self._timestamp(),
            "task_name": task.name,
            "jid": self.jid,
            "proxy_id": self.proxy_id,
            "host": host.name,
            "task_event": "completed",
            "task_type": "subtask",
            "status": "FAILED" if task.results.failed else "PASSED",
            "function": self.function,
        }
        self._emit_event(tag, data)



"""
DictInventory plugin
####################

DictInventory is an inventory plugin that loads data from Python dictionaries.
  
DictInventory sample usage
==========================

Need to instruct Nornir to use DictInventory plugin on instantiation::

    import yaml
    from nornir import InitNornir
    
    inventory_data = '''
    hosts:
      R1:
        hostname: 192.168.1.151
        platform: ios
        groups: [lab]
      R2:
        hostname: 192.168.1.153
        platform: ios
        groups: [lab]
      R3:
        hostname: 192.168.1.154
        platform: ios
        groups: [lab]
    
    groups:
      lab:
        username: cisco
        password: cisco
    '''
    
    inventory_dict = yaml.safe_load(inventory_data)
    
    NornirObj = InitNornir(
        inventory={
            "plugin": "DictInventory",
            "options": {
                "hosts": inventory_dict["hosts"],
                "groups": inventory_dict["groups"],
                "defaults": inventory_dict.get("defaults", {})
            }
        }
    )

DictInventory reference
=======================

.. autoclass:: nornir_salt.plugins.inventory.DictInventory.DictInventory
"""
import logging
from typing import Any, Dict, Type

from nornir.core.inventory import (
    Inventory,
    Group,
    Host,
    Defaults,
    ConnectionOptions,
    HostOrGroup,
    ParentGroups,
)

__version__ = "0.0.1"

logger = logging.getLogger(__name__)


def _get_connection_options(data: Dict[str, Any]) -> Dict[str, ConnectionOptions]:
    cp = {}
    for cn, c in data.items():
        cp[cn] = ConnectionOptions(
            hostname=c.get("hostname"),
            port=c.get("port"),
            username=c.get("username"),
            password=c.get("password"),
            platform=c.get("platform"),
            extras=c.get("extras"),
        )
    return cp


def _get_defaults(data: Dict[str, Any]) -> Defaults:
    return Defaults(
        hostname=data.get("hostname"),
        port=data.get("port"),
        username=data.get("username"),
        password=data.get("password"),
        platform=data.get("platform"),
        data=data.get("data"),
        connection_options=_get_connection_options(data.get("connection_options", {})),
    )


def _get_inventory_element(
    typ: Type[HostOrGroup], data: Dict[str, Any], name: str, defaults: Defaults
) -> HostOrGroup:
    return typ(
        name=name,
        hostname=data.get("hostname"),
        port=data.get("port"),
        username=data.get("username"),
        password=data.get("password"),
        platform=data.get("platform"),
        data=data.get("data"),
        groups=data.get(
            "groups"
        ),  # this is a hack, we will convert it later to the correct type
        defaults=defaults,
        connection_options=_get_connection_options(data.get("connection_options", {})),
    )


class DictInventory:
    """
    DictInventory class to instantiate inventory plugin from.

    :param hosts: dictionary with hosts data
    :param groups: dictionary with groups data
    :param defaults: dictionary with defaults data
    """

    def __init__(self, hosts: Dict, groups: Dict, defaults: Dict) -> None:
        self.hosts = hosts
        self.groups = groups
        self.defaults = defaults

    def load(self) -> Inventory:
        nr_defaults = _get_defaults(self.defaults)

        nr_hosts = {
            n: _get_inventory_element(Host, h, n, nr_defaults)
            for n, h in self.hosts.items()
        }

        nr_groups = {
            n: _get_inventory_element(Group, g, n, nr_defaults)
            for n, g in self.groups.items()
        }

        for h in nr_hosts.values():
            h.groups = ParentGroups([nr_groups[g] for g in h.groups])
        for g in nr_groups.values():
            g.groups = ParentGroups([nr_groups[g] for g in g.groups])
        return Inventory(hosts=nr_hosts, groups=nr_groups, defaults=nr_defaults)



"""
ncclient_call
#############

`Ncclient <https://github.com/ncclient>`_ is a popular library to interact with
devices using NETCONF, this plugin is a wrapper around ncclient connection 
manager object.

NETCONF protocol has a specific set of RPC calls available for use, rather
than coding separate task for each of them, ``ncclient_call`` made to execute
any arbitrary method supported by manager object plus a set of additional
helper methods to extend Ncclient library functionality.

ncclient_call sample usage
==========================

Sample code to run ``ncclient_call`` task::

    from nornir import InitNornir
    from nornir_salt import ncclient_call

    nr = InitNornir(config_file="config.yaml")
    
    output = nr.run(
        task=ncclient_call,
        call="get_config",
        source="running"
    )

ncclient_call returns
=====================

Returns XML text string by default, but can return XML data transformed
in JSON, YAML or Python format.

ncclient_call reference
=======================

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call.ncclient_call

additional methods reference
============================

ncclient_call - dir
-------------------
.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_dir

ncclient_call - help
--------------------
.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_help

ncclient_call - server_capabilities
-----------------------------------

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_server_capabilities

ncclient_call - connected
-------------------------

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_connected

ncclient_call - locked
----------------------

.. autofunction:: nornir_salt.plugins.tasks.ncclient_call._call_locked
"""
import traceback
import logging

from nornir.core.task import Result, Task
from nornir_salt.plugins.connections.NcclientPlugin import CONNECTION_NAME

log = logging.getLogger(__name__)

try:
    import lxml.etree as etree

    HAS_LXML = True
except ImportError:
    HAS_LXML = False

try:
    from ncclient.manager import OPERATIONS
    from ncclient.operations.errors import MissingCapabilityError

    HAS_NCCLIENT = True
except ImportError:
    HAS_NCCLIENT = False

try:
    # this import should work for ncclient >=0.6.10
    from ncclient.operations import GenericRPC

except ImportError:
    # for ncclient<0.6.10 need to reconstruct GenericRPC class
    if HAS_NCCLIENT:

        from ncclient.operations import RPC

        class GenericRPC(RPC):
            def request(self, data, *args, **kwargs):
                """
                :param data: (str) rpc xml string

                Testing:

                * Arista cEOS - not working, transport session closed error
                * Cisco IOS XR - working
                """
                ele = etree.fromstring(data.encode("UTF-8"))
                return self._request(ele)


def _call_locked(manager, *args, **kwargs):
    """
    Module: nornir_salt
    Task plugin: ncclient_call
    Plugin function: locked

    Helper function to run this edit-config flow:

    1. Lock target configuration/datastore
    2. Discard previous changes if any
    3. Run edit config
    4. Validate new confiugration if server supports it
    5. Run commit confirmed if server supports it
    6. Run final commit
    7. Unlock target configuration/datastore

    If any of steps 3, 4, 5, 6 fails, all changes discarded

    :param target: (str) name of datastore to edit configuration for
    :param config: (str) configuration to apply
    :param format: (str) configuration format, default is xml
    :returns result: (list) list of steps performed with details
    :returns failed: (bool) status indicator if change failed
    """
    pid = "dob04041989"
    result = []
    failed = False
    with manager.locked(target=kwargs["target"]):
        r = manager.discard_changes()
        result.append({"discard_changes": etree.tostring(r._root, pretty_print=True)})
        try:
            r = manager.edit_config(
                config=kwargs["config"],
                target=kwargs["target"],
                format=kwargs.get("format", "xml"),
            )
            result.append({"edit_config": etree.tostring(r._root, pretty_print=True)})
            # validate configuration
            try:
                r = manager.validate(source=kwargs["target"])
                result.append({"validate": etree.tostring(r._root, pretty_print=True)})
            except MissingCapabilityError:
                result.append({"validate": "MissingCapabilityError"})
                pass
            # run commit confirmed
            try:
                if kwargs.get("confirmed", True):
                    r = manager.commit(
                        confirmed=True, timeout=kwargs.get("timeout", "60"), persist=pid
                    )
                    result.append(
                        {"commit_confirmed": etree.tostring(r._root, pretty_print=True)}
                    )
                    # Could cancel but have to think about rollback criteria
                    # res = manager.cancel_commit(persist_id=pid)
            except MissingCapabilityError:
                result.append({"commit_confirmed": "MissingCapabilityError"})
                pass
            # do final commit
            r = manager.commit()
            result.append({"commit": etree.tostring(r._root, pretty_print=True)})
        except:
            tb = traceback.format_exc()
            log.error(
                "nornir_salt:ncclient_call locked edit_config call error: {}".format(tb)
            )
            result.append({"error": tb})
            failed = True
            # discard changes
            r = manager.discard_changes()
            result.append(
                {"discard_changes": etree.tostring(r._root, pretty_print=True)}
            )

    return result, failed


def _call_server_capabilities(manager, *args, **kwargs):
    """Helper function to get server capabilities"""
    return [c for c in manager.server_capabilities], False


def _call_connected(manager, *args, **kwargs):
    """Helper function to get connected status"""
    return manager.connected, False


def _call_dir(manager, *args, **kwargs):
    """Function to return alist of available methods/operations"""
    methods = (
        list(dir(manager))
        + list(manager._vendor_operations.keys())
        + list(OPERATIONS.keys())
    )
    result = sorted(
        [m for m in set(methods) if (not m.startswith("_") and not m.isupper())]
    )
    return result, False


def _call_help(manager, method_name, *args, **kwargs):
    """
    Helper function to return docstring for requested method

    :param method_name: (str) name of method or function to return docstring for
    """
    if "_call_{}".format(method_name) in globals():
        function_obj = globals()["_call_{}".format(method_name)]
    else:
        function_obj = getattr(manager, method_name)
    h = function_obj.__doc__ if hasattr(function_obj, "__doc__") else ""
    return h, False


def ncclient_call(task: Task, call: str, *args, **kwargs) -> Result:
    """
    Task to handle a call of NCClient manager object methods

    :param call: (str) ncclient manager object method to call
    :param arg: (list) any ``*args`` to use with call method
    :param kwargs: (dict) any ``**kwargs`` to use with call method
    """
    # run sanity check
    if not HAS_NCCLIENT:
        return Result(
            host=task.host,
            failed=True,
            exception="No Ncclient found, is it installed?",
        )

    # initiate parameters
    failed = False
    task.name = call

    # get rendered data if any
    if "__task__" in task.host.data:
        kwargs.update(task.host.data["__task__"])

    # check if filter formed properly - as per
    # https://ncclient.readthedocs.io/en/latest/manager.html#filter-params
    # filter should be a tuple of (type, criteria)
    if kwargs.get("filter"):
        if isinstance(kwargs["filter"], list):
            kwargs["filter"] = tuple(kwargs["filter"])
        elif isinstance(kwargs["filter"], str):
            kwargs["filter"] = tuple([kwargs.pop("ftype", "subtree"), kwargs["filter"]])

    # get Ncclient NETCONF connection object
    manager = task.host.get_connection(CONNECTION_NAME, task.nornir.config)

    # add generic RPC operation to Ncclient manager object to support RPC call
    manager._vendor_operations.update(rpc=GenericRPC)

    log.debug(
        "nornir_salt:ncclient_call calling '{}' with args: '{}'; kwargs: '{}'".format(
            call, args, kwargs
        )
    )

    # check if need to call one of helper function
    if "_call_{}".format(call) in globals():
        result, failed = globals()["_call_{}".format(call)](manager, *args, **kwargs)
    # call manager object method otherwise
    else:
        result = getattr(manager, call)(*args, **kwargs)

    # format results
    if hasattr(result, "_root"):
        result = etree.tostring(result._root, pretty_print=True).decode()
    elif isinstance(result, (list, dict, bool)):
        pass
    else:
        result = str(result)

    return Result(host=task.host, result=result, failed=failed)



"""
RetryRunner plugin
##################

RetryRunner plugin implements retry logic to improve task execution reliability.

.. warning:: For grouped tasks need to explicitly provide `connection_name` attribute
    such as `netmiko`, `napalm`, `scrapli`. Specifying `connection_name` attribute for
    standalone tasks not required. Lack of `connection_name` attribute will result in skipping
    connection retry logic and connections to all hosts initiated simultaneously up to the
    number of `num_workers`

RetryRunner Architecture
========================

.. image:: ../_images/RetryRunner_v0.png

RetryRunner Sample Usage
========================

Need to instruct Nornir to use RetryRunner on instantiation::

    from nornir import InitNornir

    NornirObj = InitNornir(
        runner={
            "plugin": "RetryRunner",
            "options": {
                "num_workers": 100,
                "num_connectors": 10,
                "connect_retry": 3,
                "connect_backoff": 1000,
                "connect_splay": 100,
                "task_retry": 3,
                "task_backoff": 1000,
                "task_splay": 100
            }
        }
    )

Sample code to demonstrate usage of ``RetryRunner``, ``DictInventory`` and ``ResultSerializer`` plugins::

    import yaml
    import pprint
    from nornir import InitNornir
    from nornir.core.task import Result, Task
    from nornir_netmiko import netmiko_send_command, netmiko_send_config
    from nornir_salt.plugins.functions import ResultSerializer

    inventory_data = '''
    hosts:
      R1:
        hostname: 192.168.1.151
        platform: ios
        groups: [lab]
      R2:
        hostname: 192.168.1.153
        platform: ios
        groups: [lab]
      R3:
        hostname: 192.168.1.154
        platform: ios
        groups: [lab]

    groups:
      lab:
        username: cisco
        password: cisco
    '''

    inventory_dict = yaml.safe_load(inventory_data)

    NornirObj = InitNornir(
        runner={
            "plugin": "RetryRunner",
            "options": {
                "num_workers": 100,
                "num_connectors": 10,
                "connect_retry": 3,
                "connect_backoff": 1000,
                "connect_splay": 100,
                "task_retry": 3,
                "task_backoff": 1000,
                "task_splay": 100
            }
        },
        inventory={
            "plugin": "DictInventory",
            "options": {
                "hosts": inventory_dict["hosts"],
                "groups": inventory_dict["groups"],
                "defaults": inventory_dict.get("defaults", {})
            }
        },
    )

    def _task_group_netmiko_send_commands(task, commands):
        # run commands
        for command in commands:
            task.run(
                task=netmiko_send_command,
                command_string=command,
                name=command
            )
        return Result(host=task.host)

    # run single task
    result1 = NornirObj.run(
        task=netmiko_send_command,
        command_string="show clock"
    )

    # run grouped tasks
    result2 = NornirObj.run(
        task=_task_group_netmiko_send_commands,
        commands=["show clock", "show run | inc hostname"],
        connection_name="netmiko"
    )

    # run another single task
    result3 = NornirObj.run(
        task=netmiko_send_command,
        command_string="show run | inc hostname"
    )

    NornirObj.close_connections()

    # Print results
    formed_result1 = ResultSerializer(result1, add_details=True)
    pprint.pprint(formed_result1, width=100)

    formed_result2 = ResultSerializer(result2, add_details=True)
    pprint.pprint(formed_result2, width=100)

    formed_result3 = ResultSerializer(result3, add_details=True)
    pprint.pprint(formed_result3, width=100)

RetryRunner - Connect to hosts behind jumphost
==============================================

RetryRunner implements logic to connect with hosts behind bastion/jumphosts.

To connect to devices behind jumphost, need to define jumphost parameters in host's inventory data::

    hosts:
      R1:
        hostname: 192.168.1.151
        platform: ios
        username: test
        password: test
        data:
          jumphost:
            hostname: 10.1.1.1
            port: 22
            password: jump_host_password
            username: jump_host_user

.. note:: Only Netmiko `connection_name="netmiko"` and Ncclient `connection_name="ncclient"`
    tasks, support connecting to hosts behind Jumphosts using above inventory data.
    
RetryRunner Reference
=====================

.. autoclass:: nornir_salt.plugins.runners.RetryRunner.RetryRunner
"""
import threading
import queue
import logging
import time
import random
from typing import List
from nornir.core.task import AggregatedResult, Task
from nornir.core.inventory import Host

try:
    import paramiko

    HAS_PARAMIKO = True
except ImportError:
    HAS_PARAMIKO = False

log = logging.getLogger(__name__)
LOCK = threading.Lock()


def worker(
    stop_event,
    connectors_q,
    work_q,
    task_backoff: int,
    task_splay: int,
    task_retry: int,
    connect_retry: int,
    reconnect_on_fail: bool,
):
    while not stop_event.is_set():
        try:
            work = work_q.get(block=True, timeout=0.1)
        except queue.Empty:
            continue
        task, host, params, result = work
        # check if need backoff task retry for this host
        if params["task_retry"] > 0:
            elapsed = time.time() - params["timestamp"]
            should_wait = (params["task_retry"] * task_backoff) / 1000
            if elapsed < should_wait:
                work_q.put(work)
                work_q.task_done()
                continue
        log.info(
            "nornir_salt:RetryRunner {} - running task '{}'".format(
                host.name, task.name
            )
        )
        time.sleep(random.randrange(0, task_splay) / 1000)
        work_result = task.start(host)
        if task.results.failed:
            log.error(
                "nornir_salt:RetryRunner {} - task execution retry attempt {} failed: '{}'".format(
                    host.name, params["task_retry"], work_result[0].exception
                )
            )
            if (
                params["task_retry"] < task_retry
                and params["connection_retry"] < connect_retry
            ):
                params["task_retry"] += 1
                params["timestamp"] = time.time()
                # recover task results for not to count task as failed
                for r in task.results:
                    r.failed = False
                    r.exception = None
                    r.skip_results = (
                        True  # for ResultSerializer to skip failed attempts
                    )
                if reconnect_on_fail:
                    # close host connections to retry them
                    close_host_connection(host, params["connection_name"])
                    connectors_q.put(work)
                    params["connection_retry"] += 1
                else:
                    work_q.put(work)
                work_q.task_done()
                continue
        # enreach result objects with runner statistics
        for result_item in work_result:
            result_item.connection_retry = params["connection_retry"]
            result_item.task_retry = params["task_retry"]
        with LOCK:
            result[host.name] = work_result
            del work_result
        work_q.task_done()
        log.info(
            "nornir_salt:RetryRunner {} - task '{}' completed".format(
                host.name, task.name
            )
        )


def connector(
    stop_event,
    connectors_q,
    work_q,
    connect_backoff: int,
    connect_splay: int,
    connect_retry: int,
    jumphosts_connections,
):
    while not stop_event.is_set():
        try:
            connection = connectors_q.get(block=True, timeout=0.1)
        except queue.Empty:
            continue
        task, host, params, result = connection
        # check if need back off connection retry for this host
        if params["connection_retry"] > 0:
            elapsed = time.time() - params["timestamp"]
            should_wait = (params["connection_retry"] * connect_backoff) / 1000
            if elapsed < should_wait:
                connectors_q.put(connection)
                connectors_q.task_done()
                continue
        # initiate connection to host
        connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
        connection_name = task.params.pop("connection_name", connection_name)
        # on connect retry get connection name from params
        connection_name = params.get("connection_name", connection_name)
        params.setdefault("connection_name", connection_name)
        if connection_name and connection_name not in host.connections:
            try:
                time.sleep(random.randrange(0, connect_splay) / 1000)
                if host.get("jumphost") and connection_name in ["netmiko", "ncclient"]:
                    extras = host.get_connection_parameters(connection_name).extras
                    extras["sock"] = connect_to_device_behind_jumphost(
                        host, jumphosts_connections
                    )
                    host.open_connection(
                        connection_name,
                        configuration=task.nornir.config,
                        extras=extras,
                    )
                else:
                    host.open_connection(
                        connection_name, configuration=task.nornir.config
                    )
                log.info(
                    "nornir_salt:RetryRunner {} - started connection: '{}'".format(
                        host.name, connection_name
                    )
                )
            except Exception as e:
                # close host connections to retry them
                close_host_connection(host, connection_name)
                log.error(
                    "nornir_salt:RetryRunner {} - connection retry attempt {}, error: '{}'".format(
                        host.name, params["connection_retry"], e
                    )
                )
                if params["connection_retry"] < connect_retry:
                    params["connection_retry"] += 1
                    params["timestamp"] = time.time()
                    connectors_q.put(connection)
                    connectors_q.task_done()
                    continue
        connectors_q.task_done()
        work_q.put((task, host, params, result))


def close_host_connection(host, connection_name):
    try:
        host.close_connection(connection_name)
    except:
        _ = host.connections.pop(connection_name, None)
    if host.get("jumphost"):
        channel_name = "jumphost_{}_channel".format(host["jumphost"]["hostname"])
        try:
            host.close_connection(channel_name)
        except:
            _ = host.connections.pop(channel_name, None)


def connect_to_device_behind_jumphost(host, jumphosts_connections):
    """
    Establish connection to devices behind jumphost/bastion
    """

    jumphost = {"timeout": 3, "look_for_keys": False, "allow_agent": False}
    jumphost.update(host["jumphost"])
    # Initiate connection to jumphost if not initiated already or
    # if it failed before
    #
    # jumphosts_connections dictionary shared between threads,
    # first thread to start will detect that no connection to
    # jumphost exists and will try to establish it, while doing
    # so, jumphost key will be added to jumphosts_connections
    # dictionary with '__connecting__' value, next threads
    # will see '__connecting__' as value for jumphost and will
    # continue sleeping until either connection succeeded or
    # failed. On connection failure, first thread will set value
    # to '__failed__' to signal other threads to exit.
    if (
        jumphost["hostname"] not in jumphosts_connections
        or jumphosts_connections.get(jumphost["hostname"]) == "__failed__"
    ):
        try:
            with LOCK:
                jumphosts_connections[jumphost["hostname"]] = "__connecting__"
            jumphost_ssh_client = paramiko.client.SSHClient()
            jumphost_ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            jumphost_ssh_client.connect(**jumphost)
            with LOCK:
                jumphosts_connections[jumphost["hostname"]] = {
                    "jumphost_ssh_client": jumphost_ssh_client,
                    "jumphost_ssh_transport": jumphost_ssh_client.get_transport(),
                }
            # add jumphost to host connections to close it
            # on nornir.close_connections() call
            host.connections[
                "jumphost_{}".format(jumphost["hostname"])
            ] = jumphost_ssh_client
            log.info(
                "nornir_salt:RetryRunner Started connection to jumphost '{}' - '{}'".format(
                    jumphost["hostname"], jumphost_ssh_client
                )
            )
        except Exception as e:
            with LOCK:
                jumphosts_connections[jumphost["hostname"]] = "__failed__"
            # add exception info to host data to include in results
            error_msg = "nornir_salt:RetryRunner failed connection to jumphost '{}', error - {}".format(
                host["jumphost"]["hostname"], e
            )
            host["exception"] = error_msg
            log.error(error_msg)
            return
    else:
        # sleep random time waiting for connection to jumphost to establish
        while jumphosts_connections[jumphost["hostname"]] == "__connecting__":
            time.sleep(random.randrange(0, 500) / 1000)
        if jumphosts_connections[jumphost["hostname"]] == "__failed__":
            # add exception info to host data to include in results
            error_msg = "nornir_salt:RetryRunner failed connection to jumphost '{}' in another thread".format(
                host["jumphost"]["hostname"]
            )
            host["exception"] = error_msg
            log.error(error_msg)
            return
    # connect to host
    channel_name = "jumphost_{}_channel".format(jumphost["hostname"])
    if not host.connections.get(channel_name):
        dest_addr = (host.hostname, host.port or 22)
        channel = jumphosts_connections[jumphost["hostname"]][
            "jumphost_ssh_transport"
        ].open_channel(
            kind="direct-tcpip",
            dest_addr=dest_addr,
            src_addr=("localhost", 7777),
            timeout=3,
        )
        host.connections[channel_name] = channel
        log.info(
            "nornir_salt:RetryRunner {} - started new channel via jumphost '{}' - '{}'".format(
                host.name,
                jumphost["hostname"],
                jumphosts_connections[jumphost["hostname"]]["jumphost_ssh_client"],
            )
        )
    return host.connections[channel_name]


class RetryRunner:
    """
    RetryRunner is a modification of QueueRunner that strives to
    make task execution as reliable as possible.

    Arguments:
        num_workers: number of threads for tasks execution
        num_connectors: number of threads for device connections
        connect_retry: number of connection attempts
        connect_backoff: exponential backoff timer in milliseconds
        connect_splay: random interval between 0 and splay for each connection in milliseconds
        task_retry: number of attempts to run task
        task_backoff: exponential backoff timer in milliseconds
        task_splay: random interval between 0 and splay before task start in milliseconds
        reconnect_on_fail: boolean, default True, perform reconnect to host on task failure
        task_timeout: int, seconds to wait for task to complete before closing all queues
            and stopping connectors and workers threads, default 600
    """

    def __init__(
        self,
        num_workers: int = 100,
        num_connectors: int = 20,
        connect_retry: int = 3,
        connect_backoff: int = 5000,
        connect_splay: int = 100,
        task_retry: int = 1,
        task_backoff: int = 5000,
        task_splay: int = 100,
        reconnect_on_fail: bool = True,
        task_timeout: int = 600,
    ) -> None:
        self.num_workers = num_workers
        self.num_connectors = num_connectors
        self.connect_retry = connect_retry
        self.connect_backoff = connect_backoff
        self.connect_splay = connect_splay
        self.task_retry = task_retry
        self.task_backoff = task_backoff
        self.task_splay = task_splay
        self.jumphosts_connections = {}
        self.reconnect_on_fail = reconnect_on_fail
        self.task_timeout = task_timeout

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        connectors_q = queue.Queue()
        work_q = queue.Queue()
        stop_event = threading.Event()
        task_timeout_event = threading.Event()
        task_timeout_timer = threading.Timer(
            self.task_timeout, lambda: task_timeout_event.set()
        )
        result = AggregatedResult(task.name)
        connector_threads = []
        worker_threads = []
        # enqueue hosts in connectors queue
        for host in hosts:
            connectors_q.put(
                (task.copy(), host, {"connection_retry": 0, "task_retry": 0}, result)
            )
        # start connectors threads
        for i in range(self.num_connectors):
            t = threading.Thread(
                target=connector,
                args=(
                    stop_event,
                    connectors_q,
                    work_q,
                    self.connect_backoff,
                    self.connect_splay,
                    self.connect_retry,
                    self.jumphosts_connections,
                ),
            )
            t.start()
            connector_threads.append(t)
        # start worker threads
        for i in range(self.num_workers):
            t = threading.Thread(
                target=worker,
                args=(
                    stop_event,
                    connectors_q,
                    work_q,
                    self.task_backoff,
                    self.task_splay,
                    self.task_retry,
                    self.connect_retry,
                    self.reconnect_on_fail,
                ),
            )
            t.start()
            worker_threads.append(t)
        # wait until all hosts completed task or timeout reached
        task_timeout_timer.start()
        while not task_timeout_event.is_set():
            with LOCK:
                hosts_no_result = [h.name for h in hosts if h.name not in result]
            if hosts_no_result == []:
                task_timeout_timer.cancel()
                break
            time.sleep(0.1)
        else:
            log.warning(
                "nornir_salt:RetryRunner task '{}', '{}'s task_timeout reached, hosts no results '{}'".format(
                    task.name, self.task_timeout, hosts_no_result
                )
            )
        # block until all queues empty
        connectors_q.join()
        work_q.join()
        # stop and delete connectors and workers threads
        stop_event.set()
        while connector_threads:
            _ = connector_threads.pop().join()
        while worker_threads:
            _ = worker_threads.pop().join()
        return result



from .ResultSerializer import ResultSerializer
from .FFun import FFun
from .HostsKeepalive import HostsKeepalive
from .TabulateFormatter import TabulateFormatter
from .DumpResults import DumpResults

__all__ = (
    "ResultSerializer",
    "FFun",
    "HostsKeepalive",
    "TabulateFormatter",
    "DumpResults",
)



"""
netmiko_send_config
###################

This task plugin relies on ``Netmiko`` conection ``send_config_set`` method
to send configuration commands to devices over SSH or Telnet.

This task plugin applies device configuration following this sequence:

- Retrieve and use, if any, per-host configuration rendered by SaltStack from host's
  inventory data ``task.host.data["__task__"]["commands"]`` or 
  ``task.host.data["__task__"]["filename"]`` locations, use configuration provided
  by ``config`` argument otherwise
- If configuration is a multi-line string, split it to a list of commands
- Check if device in enable mode, if not enter device enabled mode if device supports it
- Push configuration commands to device using ``send_config_set`` Netmiko connection's method,
  if ``batch`` argument given, pushes commands in batches
- If ``commit`` argument provided, perform configuration commit if device supports it
- If ``commit_final_delay`` argument provided, wait for a given timer and perform final commit
- Exit device configuration mode and return configuration results

Dependencies:

* `nornir-netmiko module <https://pypi.org/project/nornir-netmiko/>`_ required

netmiko_send_config sample usage
================================

Code to invoke ``netmiko_send_config`` task::

    from nornir_salt import netmiko_send_config

    output = nr.run(
        task=netmiko_send_config,
        commands=["sinterface loopback 0", "description 'configured by script'"]
    )

netmiko_send_config returns
===========================

Returns Nornir results object with individual tasks names set
equal to commands sent to device.

netmiko_send_config reference
=============================

.. autofunction:: nornir_salt.plugins.tasks.netmiko_send_config.netmiko_send_config
"""
import logging
import traceback
import time
from nornir.core.task import Result, Task

try:
    from nornir_netmiko.tasks import netmiko_send_config as nornir_netmiko_send_config

    HAS_NETMIKO = True
except ImportError:
    HAS_NETMIKO = False

log = logging.getLogger(__name__)


def netmiko_send_config(
    task: Task,
    config=None,
    commit=True,
    commit_final_delay=0,
    batch=0,
    enable=True,
    **kwargs
):
    """
    Salt-nornir Task function to send configuration to devices using
    ``nornir_netmiko.tasks.netmiko_send_config`` plugin.

    :param kwargs: (dict) any additional arguments to use with
        ``nornir_netmiko.tasks.netmiko_send_config`` task plugin
    :param config: (str or list) configuration string or list of commands to send to device
    :param commit: (bool or dict) by default commit is ``True``, as a result host
        connection commit method will be called. If ``commit`` argument is a
        dictionary, it will be supplied to connection's commit method call as ``**commit``.
    :param commit_final_delay: (int) time to wait before doing final commit, can be used in
        conjunction with commit confirm feature if device supports it.
    :param kwargs: (dict) any additional ``**kwargs`` for ``netmiko_send_config`` function.
    :param batch: (int) commands count to send in batches, sends all at once by default
    :param enable: (bool) if True (default), attempts to enter enable-mode
    :return result: Nornir result object with task execution results

    Default parameters supplied to ``netmiko_send_config`` function call::

        cmd_verify: False
        exit_config_mode: False if batch provided else unmodified

    Batch mode controlled by ``batch`` parameter, by default all configuration commands send
    at once, but that approach might lead to Netmio timeout errors if device takes too long
    to respond, sending commands in batches helps to overcome that problem.
    """
    # run sanity check
    if not HAS_NETMIKO:
        return Result(
            host=task.host,
            failed=True,
            exception="No nornir_netmiko found, is it installed?",
        )
    if kwargs.get("dry_run"):
        raise ValueError("netmiko_send_config does not support dry_run")

    task.name = "netmiko_send_config"
    kwargs.setdefault("cmd_verify", False)
    commit_final_delay = int(commit_final_delay)
    batch = max(0, int(batch))
    task_result = Result(host=task.host, result=[], changed=True)
    conn = task.host.get_connection("netmiko", task.nornir.config)

    # get configuration from host data if any
    if "commands" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["commands"]
    elif "filename" in task.host.data.get("__task__", {}):
        config = task.host.data["__task__"]["filename"]

    # transform configuration to a list if string given
    if isinstance(config, str):
        config = config.splitlines()

    # enter enable mode
    if enable and conn.check_enable_mode() is False:
        conn.enable()

    # push config to device in batches
    if batch:
        kwargs["exit_config_mode"] = False
        for i in range(0, len(config), batch):
            chunk = config[i : i + batch]
            task_result.result.append(
                conn.send_config_set(config_commands=chunk, **kwargs)
            )
        task_result.result = "\n".join(task_result.result)
    # push config all at once
    else:
        task_result.result = conn.send_config_set(config_commands=config, **kwargs)

    # check if need to commit
    if commit:
        commit = commit if isinstance(commit, dict) else {}
        try:
            conn.commit(**commit)
            # check if need to do second commit
            if commit_final_delay:
                time.sleep(commit_final_delay)
                conn.commit()
        except AttributeError:
            pass
        except:
            tb = traceback.format_exc()
            log.error("nornir-salt:netmiko_send_config commit error\n{}".format(tb))
            task_result.failed = True
            task_result.exception = tb
            task_result.chaned = False

    # check if need to exit configuration mode
    if conn.check_config_mode():
        conn.exit_config_mode()

    return task_result



"""
http_call
#########

This task plugin is to interact with devices over HTTP and is capable
of calling any method of Python built-in ``requests`` library.

http_call sample usage
======================

Sample code to run ``http_call`` task::

    from nornir_salt import http_call
    from nornir import InitNornir

    nr = InitNornir(config_file="nornir_config.yaml")

    res = nr.run(
        task=http_call,
        method="get",
        url="https://google.com",
    )

http_call returns
=================

Returns requests result string in XML or JSON format.

http_call reference
===================

.. autofunction:: nornir_salt.plugins.tasks.http_call.http_call
"""
import logging
import json
import requests
from nornir.core.task import Result, Task
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

log = logging.getLogger(__name__)

# define connection name for RetryRunner to properly detect it using:
# connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
CONNECTION_NAME = "http"


def http_call(task: Task, method: str, url: str = None, **kwargs) -> Result:
    """
    Task function to call one of the supported ``requests`` library methods.

    :param method: (str) requests method to call
    :param url: (str) relative to base_url or absolute URL to send request to
    :param kwargs: (dict) any ``**kwargs`` to use with requests method call
    :return: (str) attempt to return JSON formatted string if ``json`` string
        pattern found in response's ``Content-type`` header, returns response
        text otherwise

    ``http_call`` follows these rules to form URL to send request to:

    1. Use ``url`` attribute if provided and it absolute - starts with ``http://`` or ``https://``
    2. If ``url`` attribute provided but is relative - does not starts with ``http://`` or ``https://`` -
       it is merged with inventory ``base_url`` using formatter ``{base_url}/{url}`` if ``base_url`` is
       absolute - starts with ``http://`` or ``https://``
    3. If ``url`` attribute provided and ``base_url`` listed in inventory and they are both relative -
       does not startswith ``http://`` or ``https://`` - target URL formed using formatter:
       ``{transport}://{hostname}:{port}/{base_url}/{url}``
    4. If ``url`` attribute provided and is relative and no ``base_url`` defined in inventory ``extras``
       sections, http_call uses this formatter ``{transport}://{hostname}:{port}/{url}`` to form final URL
       if ``transport`` parameter defined in inventory ``extras`` sections
    5. If no ``url`` attribute provided use ``base_url`` to send the request if ``base_url`` defined in
       inventory ``extras`` sections and is absolute - starts with ``http://`` or ``https://``
    6. If no ``url`` attribute provided and no ``base_url`` defined in inventory
        ``extras`` sections use this formatter ``{transport}://{hostname}:{port}/``
        if ``transport`` parameter defined in inventory ``extras`` sections
    7. Raise error, unable to form URL

    Default headers added to HTTP request::

            'Content-Type': 'application/yang-data+json',
            'Accept': 'application/yang-data+json'

    If no ``auth`` attribute provided in task ``kwargs``, host's username and
    password inventory parameters used to form ``auth`` tuple.
    """
    result = None
    task.name = method

    # get rendered data if any
    if "__task__" in task.host.data:
        kwargs.update(task.host.data["__task__"])

    # get http connection inventory parameters
    conn = task.host.get_connection("http", task.nornir.config)

    log.debug(
        "nornir_salt:http_call calling '{}' with kwargs: '{}'; connection: '{}'".format(
            method, kwargs, conn
        )
    )

    # form requests parameters
    parameters = {
        "timeout": (5, 5),
        **conn["extras"],
        **kwargs,
    }

    # clean up parameters
    transport = parameters.pop("transport", None)
    base_url = parameters.pop("base_url", None)

    # add auth
    if "auth" not in parameters and conn.get("hostname") and conn.get("password"):
        parameters["auth"] = (conn["username"], conn["password"])
    # make sure auth is a tuple
    elif "auth" in parameters:
        parameters["auth"] = tuple(parameters["auth"])

    # form url
    if url:
        # if URL provided but it is relative to base url
        if not url.startswith("http://") and not url.startswith("https://"):
            # use base URL if it was provided in inventory and it is absolute
            if base_url and base_url.startswith("http"):
                parameters["url"] = "{base_url}/{url}".format(
                    base_url=base_url.strip("/"), url=url.strip("/")
                )
            # use base_url if it was provided in inventory and it is relative
            elif (
                base_url
                and not base_url.startswith("http")
                and not url.startswith(base_url)
            ):
                parameters[
                    "url"
                ] = "{transport}://{hostname}:{port}/{base_url}/{url}".format(
                    transport=transport,
                    hostname=conn["hostname"],
                    port=int(conn.get("port", 80 if transport == "http" else 443)),
                    base_url=base_url.strip("/"),
                    url=url.strip("/"),
                )
            # form URL using transport, hostname and port parameters
            elif transport:
                parameters["url"] = "{transport}://{hostname}:{port}/{url}".format(
                    transport=transport,
                    hostname=conn["hostname"],
                    port=int(conn.get("port", 80 if transport == "http" else 443)),
                    url=url.strip("/"),
                )
            else:
                raise RuntimeError(
                    "nornir-salt:http_call cannot form URL - base_url or transport required, params: {}, base_url: {}, transport: {}".format(
                        parameters, base_url, transport
                    )
                )
        else:
            parameters["url"] = url
    # use base_url if no url provided
    elif base_url and base_url.startswith("http"):
        parameters["url"] = base_url
    # form url using transport, hostname and port parameters
    elif transport:
        parameters["url"] = "{transport}://{hostname}:{port}/".format(
            transport=transport,
            hostname=conn["hostname"],
            port=int(conn.get("port", 80 if transport == "http" else 443)),
        )
    else:
        raise RuntimeError(
            "nornir-salt:http_call cannot form URL - url, base_url or transport required, params: {}, base_url: {}, transport: {}".format(
                parameters, base_url, transport
            )
        )

    # add headers
    parameters.setdefault(
        "headers",
        {
            "Content-Type": "application/yang-data+json",
            "Accept": "application/yang-data+json",
        },
    )

    # call requests method
    log.debug(
        "nornir_salt:http_call calling '{}' with params: '{}'".format(
            method, parameters
        )
    )
    response = getattr(requests, method)(**parameters)

    # form results
    if "json" in response.headers.get("Content-type", "text"):
        try:
            result = response.json()
        except requests.exceptions.ContentDecodingError:
            result = response.text
        except json.decoder.JSONDecodeError:
            result = response.text
    else:
        result = response.text

    return Result(host=task.host, result=result, failed=not response.ok)



"""
HostsKeepalive
##############

Function to iterate over Nornir hosts' connections and check if connection
still alive. If connection not responding, ``HostsKeepalive`` function deletes
it.

In general case, running ``HostsKeepalive`` function will keep connection with host
open, preventing it from timeout due to inactivity.

``HostsKeepalive`` function supports these connection types:

- netmiko - uses ``is_alive()`` method to check connection
- paramiko channel - uses connection ``conn_obj.active`` attribute to check connection status
- napalm - uses ``is_alive()`` method to check connection
- scrapli - uses ``isalive()`` method to check connection
- ncclient - uses ``connected`` attribute of connection manager to check connection status
- http - HTTP connections non-persistent hence ``HostsKeepalive`` does nothing

For other connection types ``HostsKeepalive`` logs warning message about connection 
type being unknown and keeps connection intact.

.. note:: HostsKeepalive only checks previously established connections, it
  does not creates new connections to hosts or tries to reopen dead connections.

HostsKeepalive Sample Usage
===========================

Sample code to invoke ``HostsKeepalive`` function::

    from nornir import InitNornir
    from nornir_salt.plugins.functions import HostsKeepalive

    nr = InitNornir(config_file="config.yaml")

    stats = HostsKeepalive(nr)

HostsKeepalive reference
========================

.. autofunction:: nornir_salt.plugins.functions.HostsKeepalive.HostsKeepalive
"""

import logging
import traceback

log = logging.getLogger(__name__)


def HostsKeepalive(nr):
    """
    :param nr: Nornir object
    :returns: stats dictionary with statistics about ``HostsKeepalive`` execution

    Return ``stats`` dictionary keys description:

    - ``dead_connections_cleaned`` - contains overall number of connections cleaned
    """
    stats = {"dead_connections_cleaned": 0}

    for host_name, host_obj in nr.inventory.hosts.items():
        # to avoid "RuntimeError: dictionary changed size during iteration" error
        # going to iterate over a copy of dictionary keys
        for conn_name in list(host_obj.connections.keys()):
            conn_obj = host_obj.connections[conn_name]
            is_alive = True
            try:
                if "netmiko" in str(type(conn_obj)).lower():
                    is_alive = conn_obj.connection.is_alive()
                elif "paramiko.channel.channel" in str(type(conn_obj)).lower():
                    is_alive = conn_obj.active
                elif "paramiko.client.sshclient" in str(type(conn_obj)).lower():
                    pass
                elif "napalm" in str(type(conn_obj)).lower():
                    is_alive = conn_obj.connection.is_alive()["is_alive"]
                elif "scrapli" in str(type(conn_obj)).lower():
                    is_alive = conn_obj.connection.isalive()
                elif "ncclient" in str(type(conn_obj)).lower():
                    is_alive = conn_obj.connection.connected
                elif "http" in str(type(conn_obj)).lower():
                    is_alive = True
                elif "pygnmi" in str(type(conn_obj)).lower():
                    is_alive = True
                else:
                    log.debug(
                        "nornir_salt:HostsKeepalive - uncknown connection '{}', type: '{}'".format(
                            conn_name, str(type(conn_obj)).lower()
                        )
                    )
            except:
                is_alive = False
                tb = traceback.format_exc()
                log.info(
                    "nornir_salt:HostsKeepalive - '{}' connection keepalive error: {}".format(
                        conn_name, tb
                    )
                )
            # close connection if not alive
            if not is_alive:
                host_obj.close_connection(conn_name)
                _ = host_obj.connections.pop(conn_name, None)
                stats["dead_connections_cleaned"] += 1
                log.info(
                    "nornir_salt:HostsKeepalive, removed dead '{}' connection, host '{}'".format(
                        conn_name, host_name
                    )
                )

    return stats



"""
DumpResults
###########

Function to take data and save it to the file, adding details to 
``ToFileProcessor`` and ``files`` task text database index.

``DumpResults`` does not perform any formatting on data supplied,
if it is string, it is saved as is, if it is anything but string
data converted to string using ``str(data)`` prior to saving it.

Primary use case for ``DumpResults`` is mainly related to Salt Stack
restriction on Event Bus maximum data transmission size, as a result in
certain cases need to save full results to local file system instead.
    
DumpResults sample usage
========================

Code to invoke ``DumpResults`` function ::

    from nornir import InitNornir
    from nornir_salt import TabulateFormatter
    from nornir_netmiko import netmiko_send_command
    from nornir_salt import DumpResults

    nr = InitNornir(config_file="config.yaml")

    result = NornirObj.run(
        task=netmiko_send_command,
        command_string="show clock"
    )

    serialized_output = ResultSerializer(result)
    DumpResults(serialized_output, filegroup="running_config", base_url="./tofile_outputs/")
    
DumpResults reference
=====================

.. autofunction:: nornir_salt.plugins.functions.DumpResults.DumpResults
"""

import logging
import json
import random
import os
import pprint
import time
import traceback

log = logging.getLogger(__name__)


def _load_index_data(base_url, index):
    """
    Helper function to load json index file and return python dictionary.

    :param base_url: (str) OS path to folder where to save files, default "/var/nornir-salt/"
    :param index: (str) ``ToFileProcessor`` index filename to read files information from
    :return: Dictionary of index data
    """
    index_file = os.path.join(base_url, "tf_index_{}.json".format(index))

    if not os.path.exists(index_file):
        os.makedirs(os.path.dirname(index_file), exist_ok=True)
        with open(index_file, mode="w", encoding="utf-8") as f:
            f.write(json.dumps({}))

    with open(index_file, mode="r", encoding="utf-8") as f:
        index_data = json.loads(f.read())

    return index_data


def _save_index_data(index_data, base_url, index):
    """
    Helper function to save json index file.

    :param index_data: (dict) index data to save
    :param base_url: (str) OS path to folder where to save files, default "/var/nornir-salt/"
    :param index: (str) index filename to save files information into, default is "common"
    :return: Dictionary of index data
    """
    # save new index data
    index_file = os.path.join(base_url, "tf_index_{}.json".format(index))
    with open(index_file, mode="w", encoding="utf-8") as f:
        f.write(
            json.dumps(index_data, sort_keys=True, indent=4, separators=(",", ": "))
        )


def DumpResults(
    results,
    filegroup,
    base_url="/var/nornir-salt/",
    max_files=50,
    index="common",
    proxy_id="untitled",
):
    """
    Function to save results to local file system and update ``ToFileProcessor``
    and ``files`` task text database index.

    :param results: (str, any) data to save
    :param filegroup: (str) ``tf`` file group name of files to save
    :param base_url: (str) OS path to folder where to save files, default "/var/nornir-salt/"
    :param index: (str) index name to read files index data from
    :param proxy_id: (str) filename identifier, usually equal to Proxy Minion ID
    :param max_files: (int) maximum number of files to save, older files deleted once limit reached
    :return: None

    Filename to save results formed using this formatter::

        /{base_url}/{filegroup}__{timestamp}__{rand}__{proxy_id}.txt

    Where:

    * ``base_url`` - OS path to folder where to save files
    * ``filegroup`` - ``tf`` file group name of files
    * ``timestamp`` - formed using "%d_%B_%Y_%H_%M_%S" time.strftime formatter
    * ``rand`` - random integer from 0-1000 range
    * ``proxy_id`` - Name of Proxy Minion ID
    """
    index_data = _load_index_data(base_url, index)

    # form filename
    dump_filename = "{filegroup}__{timestamp}__{rand}__{proxy_id}.txt".format(
        timestamp=time.strftime("%d_%B_%Y_%H_%M_%S"),
        rand=random.randint(0, 1000),
        proxy_id=proxy_id,
        filegroup=filegroup,
    )
    dump_filename = os.path.join(base_url, dump_filename)

    # update index data
    index_data.setdefault(filegroup, {})
    index_data[filegroup].setdefault(proxy_id, [])

    # save data to file and populate alias details for tasks
    os.makedirs(os.path.dirname(dump_filename), exist_ok=True)
    with open(dump_filename, mode="w", encoding="utf-8") as f:
        index_data[filegroup][proxy_id].insert(
            0,
            {
                "filename": dump_filename,
                "tasks": {"full_results": {}},
                "timestamp": time.strftime("%d %b %Y %H:%M:%S %Z"),
                "content_type": "pprint",
            },
        )

        # save results to file
        if isinstance(results, str):
            result_to_save = results
        else:
            result_to_save = pprint.pformat(results, indent=2, width=150)
        f.write(result_to_save + "\n")

        # update index data with span
        span = (0, len(result_to_save) + 1)
        index_data[filegroup][proxy_id][0]["tasks"]["full_results"]["span"] = span

    # check if need to delete old files
    if len(index_data[filegroup][proxy_id]) > max_files:
        file_to_rm = index_data[filegroup][proxy_id].pop(-1)
        try:
            os.remove(file_to_rm["filename"])
        except:
            log.error(
                "nornir-salt:DumpResults failed to remove file '{}':\n{}".format(
                    file_to_rm, traceback.format_exc()
                )
            )

    _save_index_data(index_data, base_url, index)



"""
connections
###########

Collection of task plugins to work with Nornir hosts' connections

Sample usage
============

Code to invoke ``onnections`` task plugins::

    TBD

API reference
=============

.. autofunction:: nornir_salt.plugins.tasks.connections.connections
.. autofunction:: nornir_salt.plugins.tasks.connections.conn_list
.. autofunction:: nornir_salt.plugins.tasks.connections.conn_close
"""
import traceback
import logging

from nornir.core.task import Result

log = logging.getLogger(__name__)


def conn_list(task, *args, **kwargs):
    """
    Function to list hos's active connections.

    return: (list) list of hosts's connections
    """
    # form list of host connections
    ret = [
        {"connection_name": conn_name, "connection_type": str(type(conn_obj))}
        for conn_name, conn_obj in task.host.connections.items()
    ]

    return Result(host=task.host, result=ret)


def conn_close(task, conn_name="all", *args, **kwargs):
    """
    Task to close host's connections.

    :param conn_name: (str) name of connection to close, default is "all"
    :return: (list) list of connections closed
    """
    ret = []

    # iterate over connections and close them
    for conn in list(task.host.connections.keys()):
        if conn_name != "all" and conn != conn_name:
            continue
        ret.append(
            {
                "connection_name": conn,
            }
        )
        try:
            task.host.close_connection(conn)
        except:
            ret[-1]["status"] = traceback.format_exc()
        _ = task.host.connections.pop(conn, None)
        ret[-1].setdefault("status", "closed")

    return Result(host=task.host, result=ret)


def connections(task, call, *args, **kwargs):
    """
    Dispatcher function to call one of the functions.

    :param call: (str) nickname of function to call
    :param arg: (list) function arguments
    :param kwargs: (dict) function key-word arguments
    :return: function execution results

    Call function nicknames:

    * ls - calls conn_list
    * close - calls conn_close
    """
    dispatcher = {
        "ls": conn_list,
        "close": conn_close,
    }
    return dispatcher[call](task, *args, **kwargs)



"""
QueueRunner plugin
##################

QueueRunner plugin implements simple queue for task execution instead of starting 
threads for ongoing tasks. 

For example, if number of threads 10, but task need to be executed on 20 hosts, 
threaded runner will start first 10 threads to run task for first 10 hosts, 
after that start another 10 threads to run task for remaining 10 hosts. 

Above process works well for majority of cases, but using QueueRunner might be
beneficial in certain situations, e.g. QueueRunner pros:

- worker threads started only once saving some negligible CPU cycles
- even if one of the hosts takes longer time to complete the task, threads will
  not stay idle and continue serving other hosts, that might reveal better
  execution time

QueueRunner Architecture
========================

.. image:: ../_images/QueueRunner_v0.png

QueueRunner Sample Usage
========================

Need to instruct Nornir to use QueueRunner on instantiation::

    from nornir import InitNornir

    NornirObj = InitNornir(
        runner={
            "plugin": "QueueRunner",
            "options": {
                "num_workers": 100
            }
        }
    )
    
QueueRunner Reference
=====================

.. autoclass:: nornir_salt.plugins.runners.QueueRunner.QueueRunner
"""
import threading
import queue
from typing import List
from nornir.core.task import AggregatedResult, Task
from nornir.core.inventory import Host


LOCK = threading.Lock()


class QueueRunner:
    """
    QueueRunner run tasks over each host using queue together
    with workers threads consuming work from work queue.

    Instead of firing up num_workes threads for each batch of hosts,
    QueueRunner starts num_workes threads once and uses queue to
    submit tasks and obtain results.

    Arguments:
        num_workers: number of threads to use
    """

    def __init__(self, num_workers: int = 20) -> None:
        self.num_workers = num_workers

    def worker(self, work_q):
        while True:
            work_to_do = work_q.get()
            if work_to_do is None:
                break
            task, host, result = work_to_do
            work_result = task.copy().start(host)
            with LOCK:
                result[host.name] = work_result
            work_q.task_done()

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        work_q = queue.Queue()
        result = AggregatedResult(task.name)
        # enqueue hosts in work queue
        for host in hosts:
            work_q.put((task, host, result))
        # start threads
        threads = []
        for i in range(self.num_workers):
            t = threading.Thread(target=self.worker, args=(work_q,), daemon=True)
            t.start()
            threads.append(t)
        # block until all tasks are done
        work_q.join()
        # stop workers:
        for i in range(self.num_workers):
            work_q.put(None)
        for t in threads:
            t.join()
        return result



"""
DataProcessor Plugin
####################

Processor plugin to process task results.

DataProcessor Sample Usage
==========================

Code to demonstrate how to use ``DataProcessor`` plugin::

    from nornir import InitNornir
    from nornir_netmiko import netmiko_send_command
    from nornir_salt import DataProcessor

    nr = InitNornir(config_file="config.yaml")

    nr_with_processor = nr.with_processors([
        DataProcessor(
            [
                {"fun": "match", "pattern": "interface.*"}
            ]
        )
    ])

    nr_with_processor.run(
        task=netmiko_send_command,
        command_string="show run"
    )

Filtering mini-query-language specification
===========================================

``lod_filter``, ``key_filter`` and ``find`` key name may be appended with check type
specifier suffix to instruct what type of check to execute with criteria against key
value in a format ``key_name__<check_type>``. For example ``key_name__glob`` would 
use glob pattern matching.

+--------------+---------------------------------------+-----------------+------------------------------+
| Check Type   | Description                           | Criteria Type   | Functions Support            |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``glob``     | Glob case sensitive pattern matching  | string          | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``re``       | Regular Expression pattern matching   | string          | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``eq``       | Equals                                | integer, string | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``ge``       | Greater than or equal to              | integer         | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``gt``       | Greater than                          | integer         | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``le``       | Less than or equal to                 | integer         | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``lt``       | Less than                             | integer         | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``in``       | Check if value in a list or in comma  | list, string    | find, lod_filter, key_filter |
|              | separated string                      |                 |                              |
+--------------+---------------------------------------+-----------------+------------------------------+
| ``contains`` | Value contains string                 | string          | find, lod_filter, key_filter |
+--------------+---------------------------------------+-----------------+------------------------------+

DataProcessor Class reference
=============================

.. autofunction:: nornir_salt.plugins.processors.DataProcessor.DataProcessor

DataProcessor Functions reference
=================================

DataProcessor Functions help to process results after task completed.

.. list-table:: DataProcessor Functions
   :widths: 15 85
   :header-rows: 1

   * - Reference Name
     - Description
   * - `find`_
     - Function to dispatch data to one of the filtering functions.
   * - `flake`_
     - Flattens python dictionary and filters its keys using `key_filter`_
   * - `flatten`_
     - Turn a nested structure into a flattened dictionary
   * - `jmespath`_
     - Query JSON string or structured data using JMESPath library
   * - `key_filter`_
     - Filter data dictionary top keys using provided patterns.
   * - `load_json`_
     - Load JSON string into python dictionary structure
   * - `load_xml`_
     - Load XML string into python dictionary structure
   * - `lod_filter`_
     - List of Dictionaries (LOD) filter function
   * - `match`_
     - Search for regex pattern in devices output
   * - `ntfsm`_
     - Pase show commands output using TextFSM ntc-templates
   * - `parse_ttp`_
     - Parses text output from device into structured data
   * - `path`_
     - Retrieves content from nested structured data at given path
   * - `run_ttp`_
     - Parse text output from device sorting results across TTP inputs
   * - `to_json`_
     - Transform Python structure to JSON formatted string
   * - `to_pprint`_
     - Transform Python structure to pprint formatted string
   * - `to_str`_
     - Transform Python structure to string without doing any formatting
   * - `to_yaml`_
     - Transform Python structure to YAML formatted string
   * - `unflatten`_
     - Turn flat dictionary produced by `flatten`_ function to a nested structure
   * - `xml_flake`_
     - Transform XML in a flattened dictionary and filter keys using `key_filter`_
   * - `xml_flatten`_
     - Transform XML in a flattened python dictionary
   * - `xml_rm_ns`_
     - Removes all namespace information from an XML Element tree
   * - `xml_to_json`_
     - Transform XML string to JSON string
   * - `xpath`_
     - Perform xpath search/filtering of XML string using LXML library

Formatter functions
-------------------

Format structured data to json, yaml etc. text string

to_str
++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.to_str

to_json
+++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.to_json

to_pprint
+++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.to_pprint

to_yaml
+++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.to_yaml

Loader functions
----------------

Load json, yaml etc. text into python structured data

load_xml
++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.load_xml

load_json
+++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.load_json

Transform functions
-------------------

Take structured data and return transformed structured data

flatten
+++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.flatten

unflatten
+++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.unflatten

xml_to_json
+++++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.xml_to_json

xml_flatten
+++++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.xml_flatten

xml_rm_ns
+++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.xml_rm_ns

path
++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.path_

Filter functions
----------------

Filter or search structured or text data

xpath
+++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.xpath

key_filter
++++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.key_filter

flake
+++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.flake

xml_flake
+++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.xml_flake

match
+++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.match

lod_filter
++++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.lod_filter

jmespath
++++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.jmespath

find
++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.find

Parse functions
---------------

Produce structured data out of text by parsing it.

ntfsm
+++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.ntfsm

parse_ttp
+++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.parse_ttp

run_ttp
+++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.run_ttp

Misc functions
--------------

Various functions with miscellaneous or utility purpose.

add_commands_from_ttp_template
++++++++++++++++++++++++++++++
.. autofunction:: nornir_salt.plugins.processors.DataProcessor.add_commands_from_ttp_template
"""
import logging
import json
import pprint
import traceback
import re
import csv
import socket

from fnmatch import fnmatchcase
from collections import deque
from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Result, Task

log = logging.getLogger(__name__)

try:
    import yaml

    HAS_YAML = True
except ImportError:
    log.debug(
        "nornir_salt:DataProcessor failed import YAML library, install: pip install pyyaml"
    )
    HAS_YAML = False

try:
    import xmltodict

    HAS_XMLTODICT = True
except ImportError:
    log.debug(
        "nornir_salt:DataProcessor failed import xmltodict library, install: pip install xmltodict"
    )
    HAS_XMLTODICT = False

try:
    from lxml import etree

    HAS_LXML = True
except ImportError:
    log.debug(
        "nornir_salt:DataProcessor failed import LXML library, install: pip install lxml"
    )
    HAS_LXML = False

try:
    from ttp import ttp

    HAS_TTP = True
except ImportError:
    log.debug(
        "nornir_salt:DataProcessor failed import TTP library, install: pip install ttp"
    )
    HAS_TTP = False

try:
    from ttp_templates import parse_output as ttp_templates_parse_output

    HAS_TTP_TEMPLATES = True
except ImportError:
    log.debug(
        "nornir_salt:DataProcessor failed import TTP templates library, install: pip install ttp_templates"
    )
    HAS_TTP_TEMPLATES = False

try:
    import jmespath as jmespath_lib

    HAS_JMESPATH = True
except ImportError:
    log.debug(
        "nornir_salt:DataProcessor failed import jmespath library, install: pip install jmespath"
    )
    HAS_JMESPATH = False

try:
    from ntc_templates.parse import parse_output as ntc_templates_parse_output

    HAS_NTFSM = True
except ImportError:
    log.debug(
        "nornir_salt:DataProcessor failed import TextFSM ntc-templates, install: pip install ntc_templates"
    )
    HAS_NTFSM = False

# --------------------------------------------------------------------------------
# formatters functions: transform structured data to json, yaml etc. text
# --------------------------------------------------------------------------------


def to_str(data, **kwargs):
    """
    Reference Name ``to_str``

    Function to transform Python structure to string without applying any formatting

    :param data: (structure) Python structure to transform
    :param kwargs: (dict) ignored
    :return: string
    """
    return str(data)


def to_json(data, **kwargs):
    """
    Reference Name ``to_json``

    Function to transform Python structure to JSON formatted string.

    :param data: (structure) Python structure to transform
    :param kwargs: (dict) additional kwargs for ``json.dumps`` method
    :return: JSON formatted string
    """
    kwargs = {"sort_keys": True, "indent": 4, "separators": (",", ": "), **kwargs}
    return json.dumps(data, **kwargs)


def to_pprint(data, **kwargs):
    """
    Reference Name ``to_pprint``

    Function to transform Python structure to pretty print string using
    ``pprint`` module

    :param data: (structure) Python structure to transform
    :param kwargs: (dict) additional kwargs for ``pprint.pformat`` method
    :return: pretty print formatted string
    """
    kwargs = {"indent": 4, **kwargs}
    return pprint.pformat(data, **kwargs)


def to_yaml(data, **kwargs):
    """
    Reference Name ``to_yaml``

    Function to transform Python structure to YAML formatted string

    Dependencies: requires PyYAML library - ``pip install pyyaml``

    :param data: (structure) Python structure to transform
    :param kwargs: (dict) additional kwargs for ``yaml.dump`` method
    :return: pretty print formatted string
    """
    kwargs = {"default_flow_style": False, **kwargs}
    if HAS_YAML:
        return yaml.dump(data, **kwargs)
    else:
        return to_pprint(data)


# --------------------------------------------------------------------------------
# loader functions: load json, yaml etc. text into python structured data
# --------------------------------------------------------------------------------


def load_xml(data, py_dict=True, **kwargs):
    """
    Reference Name ``load_xml``

    Load XML string into python dictionary structure using xmltodict library.

    Dependencies: requires LXML library - ``pip install lxml``

    :param data: (str) XML formatted string
    :param py_dict: (bool) if True (default), will transform structure returned
      by ``xmltodict`` to normal dictionary instead of ``OrderedDict``
    :param kwargs: (dict) any additional ``**kwargs`` for ``xmltodict.parse`` method
    :returns: python dictionary
    """
    parsed = xmltodict.parse(data, **kwargs)
    # pass parsed results through json to get rid of ordered dictionary
    if py_dict:
        parsed = json.loads(json.dumps(parsed))
    return parsed


def load_json(data, **kwargs):
    """
    Reference Name ``load_json``

    Load JSON string into python dictionary structure using json library.

    :param data: (str) JSON formatted string
    :param kwargs: (dict) any additional ``**kwargs`` for ``json.loads`` method
    :returns: python dictionary
    """
    return json.loads(data, **kwargs)


# --------------------------------------------------------------------------------
# transform functions: take structured data and return processes structured data
# --------------------------------------------------------------------------------


def flatten(data, parent_key="", separator=".", **kwargs):
    """
    Reference Name ``flatten``

    Turn a nested structure (combination of lists/dictionaries) into a
    flattened dictionary.

    This function is useful to explore deeply nested structures such as XML
    output obtained from devices over NETCONF.

    Another usecase is filtering of the keys in resulted dictionary, as
    they are the strings, glob or regex matching can be applied on them.

    :param data: nested data to flatten
    :param parent_key: string to prepend to dictionary's keys, used by recursion
    :param separator: string to separate flattened keys
    :return: flattened structure

    Based on Stackoverflow answer:
    https://stackoverflow.com/a/62186053/12300761

    All credits for the idea to https://github.com/ScriptSmith

    Sample usage::

        flatten({'a': 1, 'c': {'a': 2, 'b': {'x': 5, 'y' : 10}}, 'd': [1, 2, 3] })

        >> {'a': 1, 'c.a': 2, 'c.b.x': 5, 'c.b.y': 10, 'd.0': 1, 'd.1': 2, 'd.2': 3}
    """
    items = []
    if isinstance(data, dict):
        for key, value in data.items():
            new_key = "{}{}{}".format(parent_key, separator, key) if parent_key else key
            items.extend(flatten(value, new_key, separator).items())
    elif isinstance(data, list):
        for k, v in enumerate(data):
            new_key = "{}{}{}".format(parent_key, separator, k) if parent_key else k
            items.extend(flatten({str(new_key): v}).items())
    else:
        items.append((parent_key, data))
    return dict(items)


def unflatten(data, separator=".", **kwargs):
    """
    Reference Name ``unflatten``

    Turn flat dictionary produced by flatten function to a nested structure

    :param data: flattened dictionary
    :param separator: string to split flattened keys
    :return: nested structure

    List indexes must follow in order, for example this flattened dictionary::

        {
            "5.a.b.0.c": 1,
            "5.a.b.1.c": 2,
            "10.a.b.0.c": 3,
            "10.a.b.1": 4,
        }

    Will produce unexpected results due to indexes not following numerical order::

        [{'a': {'b': [{'c': 1}]}},
         {'a': {'b': [{'c': 2}]}},
         {'a': {'b': [{'c': 3}]}},
         {'a': {'b': [4]}}]}
    """
    # decide if overall structure should be a list or a dictionary
    first_path_item = list(data.keys())[0].split(separator)[0]
    if first_path_item.isdigit():
        res = []
    else:
        res = {}
    # un-flatten data structure
    for key, value in data.items():
        tracker = res
        path = key.split(separator)
        for index, item in enumerate(path[:-1]):
            if isinstance(tracker, dict):
                # if next level is a list
                if path[index + 1].isdigit():
                    tracker.setdefault(item, [])
                else:
                    tracker.setdefault(item, {})
                tracker = tracker[item]
            elif isinstance(tracker, list):
                try:
                    tracker = tracker[int(item)]
                except IndexError:
                    # if next level is a list
                    if path[index + 1].isdigit():
                        tracker.append([])
                    else:
                        tracker.append({})
                    tracker = tracker[-1]
        if isinstance(tracker, dict):
            tracker[path[-1]] = value
        elif isinstance(tracker, list):
            tracker.append(value)

    return res


def xml_to_json(data, **kwargs):
    """
    Reference name ``xml_to_json``

    Dependencies: requires LXML library - ``pip install lxml``

    Function to transform XML string to JSON string.

    Steps are:

    1. Transform XML to Python dictionary using xmltodict by calling ``load_xml`` function
    2. Serialize Python dictionary to JSON string by calling ``to_json`` function

    :param data: (str) XML formatted string
    :param kwargs: (dict) kwargs to use with ``to_json`` function
    :return: JSON formatted string
    """
    return to_json(data=load_xml(data, py_dict=False), **kwargs)


def xml_flatten(data, **kwargs):
    """
    Reference name ``xml_flatten``

    Dependencies: requires LXML library - ``pip install lxml``

    Function to transform XML in a flattened python dictionary representation

    Steps are:

    1. Transform XML to Python dictionary using xmltodict by calling ``load_xml`` function
    2. Flatten python dictionary calling ``flatten`` function

    :param data: (str) XML formatted string
    :param kwargs: (dict) kwargs to use with ``flatten`` function
    :return: flattened python dictionary
    """
    return flatten(data=load_xml(data, py_dict=False), **kwargs)


def xml_rm_ns(data, recover=True, ret_xml=True, **kwargs):
    """
    Reference Name ``xml_rm_ns``

    Namespace clean up code taken from:
    https://github.com/jeremyschulman/xml-tutorial/blob/master/strip-namespaces.md

    This function removes all namespace information from an XML Element tree
    so that a Caller can then use the `xpath` function without having
    to deal with the complexities of namespaces.

    Dependencies: requires LXML library - ``pip install lxml``

    :param data: (str) XML formatted string
    :param recover: (bool) if True (default) uses ``etree.XMLParser(recover=True)`` to
        parse XML, can help to recover bad XML
    :param ret_xml: (bool) indicates what to return, default is True
    :param kwargs: (dict) any additional ``**kwargs`` are ignored
    :returns: XML string with no namespaces if ret_xml is True, ``etree.Element`` otherwise
    """
    if recover:
        tree = etree.fromstring(data, parser=etree.XMLParser(recover=True))
    else:
        tree = etree.fromstring(data)

    # first we visit each node in the tree and set the tag name to its localname
    # value; thus removing its namespace prefix

    for elem in tree.getiterator():
        elem.tag = etree.QName(elem).localname

    # at this point there are no tags with namespaces, so we run the cleanup
    # process to remove the namespace definitions from within the tree.

    etree.cleanup_namespaces(tree)

    if ret_xml:
        return etree.tostring(tree, pretty_print=True, encoding="utf-8").decode()
    else:
        return tree


def path_(data, path, **kwargs):
    """
    Reference name ``path_``

    Function to retrieve content from nested structured data at given path.

    :param path: (str, list) dot separated path to result or list of path items
    :param data: (dict) data to get results from
    :return: results at given path

    Sample data::

        {
            "VIP_cfg": {
                "1.1.1.1": {
                    "config_state": "dis",
                    "services": {
                        "443": {
                            "https": [
                                {"real_port": "443"}
                            ],
                        }
                    }
                }
            }
        }

    With ``path`` ``"VIP_cfg.'1.1.1.1'.services.443.https.0.real_port"``
    will return ``443``
    """
    ret = data

    # form path list
    if isinstance(path, str):
        # perform path split accounting for quotes inside path
        path_list = [""]
        inside_quotes = False
        for char in path:
            if char == "." and not inside_quotes:
                path_list.append("")
            elif char in ["'", '"']:
                inside_quotes = not inside_quotes
            else:
                path_list[-1] += char
    elif isinstance(path, list):
        path_list = path
    else:
        raise TypeError(
            "nornir-salt:DataProcessor:path unsupported path type {}".format(type(path))
        )

    # descend down the data path
    for item in path_list:
        if item in ret or isinstance(item, int):
            ret = ret[item]
        elif item.isdigit():
            ret = ret[int(item)]

    return ret


# --------------------------------------------------------------------------------
# filtering functions: filter structured/text data
# --------------------------------------------------------------------------------


# check functions
def _check_glob(value, criteria):
    # returns True if glob pattern matches value
    return fnmatchcase(str(value), str(criteria))


def _check_regex(value, criteria):
    # returns True if regex pattern matches value
    return True if re.search(str(criteria), str(value)) else False


def _check_equal(value, criteria):
    # checks for equality string or integer
    try:
        return int(value) == int(criteria)
    except ValueError:
        return value == criteria


def _check_greater_or_equal(value, criteria):
    # compare integers
    try:
        return int(value) >= int(criteria)
    except ValueError:
        return False


def _check_greater_than(value, criteria):
    # compare integers
    try:
        return int(value) > int(criteria)
    except ValueError:
        return False


def _check_less_or_equal(value, criteria):
    # compare integers
    try:
        return int(value) <= int(criteria)
    except ValueError:
        return False


def _check_less_than(value, criteria):
    # compare integers
    try:
        return int(value) < int(criteria)
    except ValueError:
        return False


def _check_in(value, criteria):
    # check if value is in criteria list
    try:
        if isinstance(criteria, list):
            return value in criteria
        elif isinstance(criteria, str):
            return value in [i.strip() for i in criteria.split(",")]
        else:
            return str(value) in [str(criteria)]
    except TypeError:
        return False


def _check_contains(value, criteria):
    # check if criteria is in value string
    return str(criteria) in str(value)


check_fun_dispatcher = {
    "glob": _check_glob,
    "re": _check_regex,
    "eq": _check_equal,
    "ge": _check_greater_or_equal,
    "gt": _check_greater_than,
    "le": _check_less_or_equal,
    "lt": _check_less_than,
    "in": _check_in,
    "contains": _check_contains,
}


def _form_check_list(checks_dictionary):
    """
    Helper function to form a list of filtering checks.

    :param checks_dictionary: (dict) dictionary where keys used to indicate
        check type and value is the criteria to check. Default check type is
        glob case sensitive pattern matching.
    :return: list of dictionaries

    ``checks_dictionary`` keys can use filtering mini-query-language specifiers
    to indicate check type.

    Returns list of dictionaries::

        [
            {
                "fun": _check_fun_reference,
                "key": checks_dictionary_key_name,
                "criteria": checks_dictionary_key_value,
            }
        ]
    """
    checks = []
    for key_name, criteria in checks_dictionary.items():
        if "__" in key_name:
            key_name, *check_type = key_name.split("__")
            check_type = "__".join(check_type)
        else:
            check_type = "glob"
        if check_type in check_fun_dispatcher:
            checks.append(
                {
                    "fun": check_fun_dispatcher[check_type],
                    "key": key_name,
                    "criteria": criteria,
                }
            )
        else:
            log.error("nornir-salt:DataProcessor unknown check '{}'".format(check_type))
            return []
    return checks


def xpath(data, expr, rm_ns=False, recover=False, **kwargs):
    """
    Reference Name ``xpath``

    Function to perform xpath search/filtering of XML string using LXML library.

    Dependencies: requires LXML library - ``pip install lxml``

    :param data: (str) XML formatted string
    :param expr: (str) xpath expression to use
    :param rm_ns: (bool) default is False, if True removes namespace from XML string using
        ``xml_rm_ns`` function
    :param recover: (bool) default is False, if True uses ``etree.XMLParser(recover=True)``
        while loading XML from string
    :param kwarg: (dict) ``**kwargs`` to use for LXML etree.xpath method
    :return: XML filtered string
    """
    if not HAS_LXML:
        return data

    if rm_ns:
        tree = xml_rm_ns(data, ret_xml=False, recover=recover)
    elif recover:
        tree = etree.fromstring(data, parser=etree.XMLParser(recover=True))
    else:
        tree = etree.fromstring(data)

    filtered = tree.xpath(expr, **kwargs)

    if isinstance(filtered, list):
        res = [
            etree.tostring(i, pretty_print=True, encoding="utf-8").decode()
            for i in filtered
        ]
        return "\n".join(res)

    return etree.tostring(filtered, pretty_print=True, encoding="utf-8").decode()


def key_filter(data, pattern=None, checks_required=True, **kwargs):
    """
    Reference Name ``key_filter``

    Function to filter data dictionary top keys using provided patterns.

    :param data: (dictionary) Python dictionary
    :param kwargs: (dict) any additional kwargs are key and value pairs, where key name
        is arbitrary and used to indicate check type following `Filtering mini-query-language specification`_
        and value is the criteria to check. Default check type is glob case sensitive
        pattern matching.
    :param pattern: (str) pattern to use for filtering
    :param checks_required: (bool) if True (default) returns empty dictionary if no checks provided,
        returns all otherwise
    :return: filtered python dictionary

    Default logic is key name must pass **any** of the criteria provided.

    Sample usage::

        key_filter(
            data=data_dictionary,
            pattern="1234*",
            pattern__glob="abc*",
            pattern2__glob="*abc*",
            pattern__re="abc.*",
        )

    Filtered dictionary key name must satisfy at least one of the matching criteria.
    """
    if not isinstance(data, dict):
        log.warning(
            "nornir_salt:DataProcessor:key_filter skipping, data is not dictionary but {}".format(
                type(data)
            )
        )
        return data

    if pattern:
        kwargs["pattern"] = pattern

    checks = _form_check_list(kwargs)
    log.debug(
        "nornir_salt:DataProcessor:key_filter running filter checks {}".format(checks)
    )
    # return empty results if no checks
    if not checks and checks_required:
        return {}

    return {
        k: data[k]
        for k in data.keys()
        if any([c["fun"](k, c["criteria"]) for c in checks])
    }


def lod_filter(data, pass_all=True, strict=True, checks_required=True, **kwargs):
    """
    Reference Name ``lod_filter``

    List of Dictionaries (LOD) filter function.

    Iterates over list of dictionaries and returns dictionary items that have
    value of key matched by glob pattern.

    Patterns are case sensitive.

    Key value converted to string to perform match check.

    :param data: (list) list of dictionaries to search in
    :param kwargs: (dict) any additional kwargs are key and value pairs, where key is a name
        of the dictionary key to search for and value is the criteria to check. Default check
        type is glob case sensitive pattern matching.
    :param pass_all: (bool) if True (default) logic is AND - dictionary must pass ALL
        checks, if False logic is ANY
    :param strict: (bool) if True (default) invalidates list dictionary item
        if no criteria key found in dictionary
    :param checks_required: (bool) if True (default) returns empty list if no checks provided,
        returns all otherwise
    :return: filtered list of dictionaries
    """
    if not isinstance(data, list):
        log.warning(
            "nornir_salt:DataProcessor:lod_filter skipping, data is not list but {}".format(
                type(data)
            )
        )
        return data

    checks = _form_check_list(kwargs)
    log.debug(
        "nornir_salt:DataProcessor:lod_filter running filter checks {}".format(checks)
    )
    # return empty results if no checks
    if not checks and checks_required:
        return []

    # run filtering
    if pass_all:
        return [
            i
            for i in data
            if all(
                [
                    c["fun"](i[c["key"]], c["criteria"])
                    if c["key"] in i
                    else not strict
                    for c in checks
                ]
            )
        ]
    else:
        return [
            i
            for i in data
            if any(
                [
                    c["fun"](i[c["key"]], c["criteria"])
                    if c["key"] in i
                    else not strict
                    for c in checks
                ]
            )
        ]


def match(data, pattern, before=0, **kwargs):
    """
    Reference name ``match``

    Function to search for regex pattern in devices output, similar to network
    devices ``include/match`` pipe statements.

    :param data: multiline string to search in
    :param pattern: pattern to search for, glob (default) or regex
    :param before: number of lines before match to include in results
    :return: filtered string
    """
    # do sanity check
    if not isinstance(data, str):
        return data

    regex = re.compile(str(pattern))
    searched_result = []
    lines_before = deque([], abs(before))

    # iterate over results and search for matches
    for line in iter(data.splitlines()):
        if regex.search(line):
            searched_result.append(
                line
                if before == 0
                else "--\n{}\n{}".format("\n".join(lines_before), line)
            )
        lines_before.append(line)

    return "\n".join(searched_result)


def flake(data, **kwargs):
    """
    Reference name ``flake``

    FLAKE - Flattened Key Filter

    Function to transform Python structure in a flattened python dictionary
    representation and filter its keys using ``key_filter`` function.

    Steps are:

    1. Transform data to flattened Python dictionary using ``flatten`` function
    2. Filter Python dictionary keys using ``key_filter`` function

    :param data: (dict or list) structured data
    :param kwargs: (dict) kwargs to use with ``key_filter`` function
    :return: flattened and filtered python dictionary
    """
    return key_filter(flatten(data), **kwargs)


def xml_flake(data, pattern, **kwargs):
    """
    Reference name ``xml_flake``

    XML FLAKE - XML Flattened Key Filter

    Function to transform XML in a flattened python dictionary representation and
    filter its keys using ``key_filter`` function

    Steps are:

    1. Transform XML to flattened Python dictionary using ``xml_flatten`` function
    2. Filter Python dictionary keys using ``key_filter`` function

    :param data: (str) XML formatted string
    :param kwargs: (dict) kwargs to use with ``key_filter`` function
    :return: flattened and filtered python dictionary
    """
    return key_filter(xml_flatten(data), pattern=pattern, **kwargs)


def jmespath(data, expr, **kwargs):
    """
    Reference name ``jmespath``

    JMESPath is a query language for JSON

    Function that uses `JMESPath library <https://jmespath.org/>`_  to filter and
    search structured data.

    Requires `JMESPath package <https://pypi.org/project/jmespath/>`_ installed on the
    minion.

    :param data: (struct or str) list or dictionary structure or JSON formatted string
    :param expr: (str) jmespath query expression
    :return: (struct) query results
    """
    if not HAS_JMESPATH:
        log.error("nornir_salt:DataProcessor:jmespath failed import jmespath library")
        return data

    if isinstance(data, str):
        return jmespath_lib.search(expr, json.loads(data))
    else:
        return jmespath_lib.search(expr, data)


def find(data, path=None, use_jmespath=False, use_xpath=False, **kwargs):
    """
    Reference name ``find``

    Function to dispatch data to one of the filtering functions based on data type.

    :param data: (list, dict, str) data to search in
    :param use_jmespath: (bool) default is False, if True uses jmespath library
    :param use_xpath: (bool) default is False, if True use lxml library xpath
    :param path: (str) dot separated path or list of path items to results
        within data to search in; if ``use_jmespath`` is True, path must be
        jmespath query expression; if ``use_xpath`` is True, path must be
        lxml library xpath expression;
    :return: search results

    Searching dispatch rules:

    * if ``use_jmespath`` is True uses ``jmespath`` function with ``path`` provided
    * if ``use_xpath`` is True uses ``xpath`` function with ``path`` provided

    Next, if supplied data is a list or dictionary and ``path`` provided calls
    ``path_`` function to narrow down results before dispatching them further:

    * if result type is a list uses ``lod_filter``
    * if result type is a dictionary uses ``key_filter``
    * if result type is a string uses ``match`` function
    """
    result = data

    # use need to use external lib
    if use_jmespath and path:
        return jmespath(data, path, **kwargs)
    elif use_xpath and path:
        return xpath(data, path, **kwargs)

    # check if need to narrow down result to given path
    if path:
        result = path_(data, path)

    # use one of DataProcessor filters depending on data type
    if isinstance(result, list):
        return lod_filter(result, **kwargs)
    elif isinstance(result, dict):
        return key_filter(result, **kwargs)
    elif isinstance(result, str):
        return match(result, **kwargs)


# --------------------------------------------------------------------------------
# parsing functions: parse text data - return structured data
# --------------------------------------------------------------------------------


def parse_ttp(
    result: Result,
    task: Task,
    host,
    template: str = None,
    ttp_kwargs={},
    res_kwargs={},
):
    """
    Reference name ``parse_ttp``

    Dependencies: requires TTP library - ``pip install ttp``

    Function to parse text output from device and return structured data

    :param result: (obj) Nornir Result object
    :param task: (obj) Nornir Task object
    :param task: (obj) Nornir Host object
    :param template: (str) TTP template string or reference to ``ttp://`` templates
    :param ttp_kwargs: (dict) dictionary to use while instantiating TTP parse object
        or to source ``ttp_variables`` argument for ``ttp_templates parse_output`` method
    :param res_kwargs: (dict) dictionary to use with ``result`` method
        or to source ``structure`` argument for ``ttp_templates parse_output`` method
    :return: parsed structure

    If no template provided uses ``ttp_templates`` rpeository to source the template
    based on host platform and CLI command string. In such a case to determine cli
    command to properly form TTP template name, task/subtask that produced Result
    object must use CLI command string as a name. For example::

        from nornir import InitNornir
        from nornir_netmiko import netmiko_send_command
        from nornir_salt import DataProcessor

        nr = InitNornir(config_file="config.yaml")

        nr_with_processor = nr.with_processors([
            DataProcessor([
                {
                    "fun": "parse_ttp", "res_kwargs": {"structure": "flat_list"}
                }
            ])
        ])

        nr_with_processor.run(
            task=netmiko_send_command,
            command_string="show version",
            name="show version",
        )
    """
    if not HAS_TTP:
        log.warning("nornir_salt:DataProcessor:parse_ttp failed import TTP library")
        return

    if not isinstance(result.result, str):
        log.warning(
            "nornir_salt:DataProcessor:parse_ttp skipping, result is not string but {}".format(
                type(result.result)
            )
        )
        return

    # do parsing
    if template:
        parser = ttp(result.result, template, **ttp_kwargs)
        parser.parse(one=True)
        result.result = parser.result(**res_kwargs)
    elif HAS_TTP_TEMPLATES:
        result.result = ttp_templates_parse_output(
            data=result.result,
            platform=host.platform,
            command=result.name,
            structure=res_kwargs.get("structure", "list"),
            template_vars=ttp_kwargs.get("vars", {}),
        )
    else:
        log.error(
            "nornir_salt:DataProcessor:parse_ttp template not provided and ttp_templates not found, do nothing"
        )


def run_ttp(
    data: MultiResult,
    template: str,
    ttp_kwargs={},
    res_kwargs={},
    task: Task = None,
    remove_tasks=True,
    **kwargs,
):
    """
    Reference name ``run_ttp``

    Dependencies: requires TTP library - ``pip install ttp``

    Function to parse text output from device sorting results across TTP inputs
    based on commands values.

    :param data: (str) Nornir MultiResult object
    :param template: (str) TTP template string or reference to ``ttp://`` templates
    :param ttp_kwargs: (dict) dictionary to use while instantiating TTP parse object
    :param res_kwargs: (dict) dictionary to use with ``result`` method
    :param remove_tasks: (bool) if set to True and data is MultiResult object will remove
        other task results
    :param task: (obj) Nornir Task object, used to form parsing results
    :return: Nothing, add parsing results to provided MultiResult object

    Provided Nornir MultiResult object processed by sorting task results across TTP
    Template inputs to parse. After parsing, all other tasks' results removed from
    MultiResult object and parsing results appended to it. If ``remove_tasks`` set to
    False, other tasks results not removed.

    Sample usage::

        import pprint
        from nornir import InitNornir
        from nornir_salt import ResultSerializer, DataProcessor, netmiko_send_commands

        nr = InitNornir(config_file="nornir_config.yaml")

        # define TTP template with inputs having commands attributes
        template = '''
        <input name="arp">
        commands = ["show arp"]
        </input>

        <input name="version">
        commands = ["show version"]
        </input>

        <group name="arp_cache*" input="arp">
        Internet  {{ ip }}   {{ age }}   {{ mac }}  ARPA   {{ interface }}
        </group>

        <group name="facts.version" input="version">
        Cisco IOS XE Software, Version {{ iose_xe_version }}
        </group>
        '''

        # add data processor with run_ttp function
        nr_with_dp = nr.with_processors([DataProcessor(
            [{"fun": "run_ttp", "template": template}]
        )])

        # run task; commands for task will be dynamically populated by DataProcessor
        # run_ttp function with commands defined within TTP template inputs
        result = nr_with_dp.run(task=netmiko_send_commands)

        # serialize results to dictionary
        dict_result = ResultSerializer(result)

        pprint.pprint(dict_result)

    Running above code should print results::

        {'hostname': {'run_ttp': [[{'arp_cache': [{'age': '2',
                                    'interface': 'GigabitEthernet1',
                                    'ip': '10.10.20.28',
                                    'mac': '0050.56bf.f0be'},
                                   {'age': '-',
                                    'interface': 'GigabitEthernet1',
                                    'ip': '10.10.20.48',
                                    'mac': '0050.56bf.9379'}]},
                    {'facts': {'version': {'iose_xe_version': '16.09.03'}}}]]}}
    """
    if not HAS_TTP:
        log.warning("nornir_salt:DataProcessor:parse_ttp failed import TTP library")
        return

    if not isinstance(data, MultiResult):
        log.warning(
            "nornir_salt:DataProcessor:run_ttp skipping, data is not MultiResult but {}".format(
                type(data)
            )
        )
        return

    parser = ttp(template=template, **ttp_kwargs)
    ttp_inputs_load = parser.get_input_load()

    # go over template's inputs and add output from devices
    for template_name, inputs in ttp_inputs_load.items():
        # if no inputs defined, add all to default input
        if not inputs:
            default_input_data = []
            for i in data:
                # check if need to skip this task
                if hasattr(i, "skip_results") and i.skip_results is True:
                    continue
                if isinstance(i.result, str):
                    default_input_data.append(i.result)
            if default_input_data:
                parser.add_input(
                    data="\n".join(default_input_data),
                    input_name="Default_Input",
                    template_name=template_name,
                )
        # sort results across inputs with commands
        else:
            for input_name, input_params in inputs.items():
                commands = input_params.get("commands", [])
                input_data = []
                for i in data:
                    # check if need to skip this task
                    if hasattr(i, "skip_results") and i.skip_results is True:
                        continue
                    if i.name in commands and isinstance(i.result, str):
                        input_data.append(i.result)
                if input_data:
                    parser.add_input(
                        data="\n".join(input_data),
                        input_name=input_name,
                        template_name=template_name,
                    )

    # run parsing in single process
    parser.parse(one=True)

    # remove other task results
    if remove_tasks:
        while data:
            _ = data.pop()

    # add parsing results
    data.append(
        Result(host=task.host, result=parser.result(**res_kwargs), name="run_ttp")
    )


def ntfsm(result: Result, task: Task, host, **kwargs):
    """
    Reference name ``ntfsm``

    This function performs CLI show comamnds output parsing using
    `TextFSM ntc-templates <https://github.com/networktocode/ntc-templates>`_.

    :param result: (obj) Nornir Result object
    :param task: (obj) Nornir Task object
    :param host: (obj) Nornir Host object
    :returns: Nothing, modifies provided Result object to save parsing results

    If no such a template available to parse show command output, ntc-templates
    ``parse_output`` method return empty list.

    For this function to determine cli command to properly form TextFSM template
    name, task/subtask that produced Result object must use cli command string
    as a name. For example::

        from nornir import InitNornir
        from nornir_netmiko import netmiko_send_command
        from nornir_salt import DataProcessor

        nr = InitNornir(config_file="config.yaml")

        nr_with_processor = nr.with_processors([
            DataProcessor([{"fun": "ntfsm"}])
        ])

        nr_with_processor.run(
            task=netmiko_send_command,
            command_string="show version",
            name="show version",
        )
    """
    # do sanity checks
    if not HAS_NTFSM:
        log.warning(
            "nornir_salt:DataProcessor:ntfsm failed import ntc-templates library"
        )
        return
    if not isinstance(result, Result):
        log.warning(
            "nornir_salt:DataProcessor:ntfsm skipping, data is not Result but {}".format(
                type(result)
            )
        )
        return

    # parse output
    result.result = ntc_templates_parse_output(
        platform=host.platform,
        command=result.name,
        data=result.result,
    )


# --------------------------------------------------------------------------------
# misc
# --------------------------------------------------------------------------------


def add_commands_from_ttp_template(task, template, **kwargs):
    """
    Function to extract commands from TTP template and add them
    to task params. Used in conjunction with ``run_ttp`` parsing function.

    This function called at ``task_started`` point, allowing to update
    task's ``params`` dictionary with ``commands`` to collect from devices
    prior to further executing the task.

    Commands extracted from TTP Template inputs that may contain definitions
    of ``commands`` list.

    Dependencies: requires TTP library - ``pip install ttp``.

    :param task: (obj) Nornir task object
    :param template: (str) TTP Template string
    """
    if not HAS_TTP:
        log.warning(
            "nornir_salt:DataProcessor:extract_commands_from_ttp failed import TTP library"
        )
        return

    task.params.setdefault("commands", [])
    parser = ttp(template=template)
    ttp_inputs_load = parser.get_input_load()

    # go over template's inputs and collect commands to get from devices
    for template_name, inputs in ttp_inputs_load.items():
        for input_name, input_params in inputs.items():
            for cmd in input_params.get("commands", []):
                if cmd not in task.params["commands"]:
                    task.params["commands"].append(cmd)


def iplkp(
    data,
    use_dns=False,
    use_csv=None,
    subform="{ip}({lookup})",
    pattern_ipv4=r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}",
    pattern_ipv6=r"(?:[a-fA-F0-9]{1,4}:|:){1,7}(?:[a-fA-F0-9]{1,4}|:?)",
    **kwargs,
):
    """
    Reference name ``iplkp``

    Function to resolve IPv4 and IPv6 addresses using DNS or CSV lookups
    and substitute them in original data with lookup results.

    :param use_dns: (bool) if true use DNS to resolve IP addresses
    :param use_csv: (str) csv formatted string for lookup
    :param subform: (str) substitute formatter to use with Python ``format`` function
        to replace IP addresses, accepts two variables ``ip`` - to hold original IP
        addresses value and ``lookup`` for lookup result
    :param pattern_ipv4: (str) pattern to use to search for IPv4 addresses to replace
    :param pattern_ipv6: (str) pattern to use to search for IPv6 addresses to replace
    :return: results with IP addresses being replaced with lookup values
    """
    # run sanity checks
    if not use_dns and not use_csv:
        return data
    if not isinstance(data, str):
        return data

    ret = data
    lookup_results = {}

    # load csv data
    if use_csv:
        lookup_data = {
            row[0]: row[1]
            for row in csv.reader(iter(use_csv.splitlines()))
            if len(row) >= 2
        }

    # find all IPv4 and IPv6 matches in data
    ips_v4 = re.findall(pattern_ipv4, data)
    ips_v6 = re.findall(pattern_ipv6, data)

    # resolve IP addresses using DNS
    if use_dns:
        # lookup IPv4 addresses
        for ip in ips_v4:
            try:
                lookup_results[ip] = socket.gethostbyaddr(ip)[0]
            except socket.herror:
                continue
        # lookup IPv6 addresses
        for ip in ips_v6:
            try:
                lookup_results[ip] = socket.gethostbyaddr(ip)[0]
            except socket.herror:
                continue

    # resolve IP addresses using CSV data
    if use_csv:
        # lookup IPv4 addresses
        lookup_results.update(
            {ip: lookup_data[ip] for ip in ips_v4 if ip in lookup_data}
        )
        # lookup IPv6 addresses
        lookup_results.update(
            {ip: lookup_data[ip] for ip in ips_v6 if ip in lookup_data}
        )

    # replace IP addresses in original data
    for ip, lookup in lookup_results.items():
        # use negative lookahead to avoid "1.1.1.1" matching "1.1.1.11"
        ret = re.sub(
            r"{}(?!\d+)".format(re.escape(ip)),
            subform.format(ip=ip, lookup=lookup),
            ret,
        )

    return ret


# --------------------------------------------------------------------------------
# functions dispatcher dictionary
# --------------------------------------------------------------------------------

task_instance_completed_dispatcher_per_result_data = {
    # functions in this dictionary get task result data and kwargs
    #
    # formatters - structured data to text
    "to_str": to_str,
    "to_json": to_json,
    "to_pprint": to_pprint,
    "to_yaml": to_yaml,
    # loaders - text to structured data
    "load_xml": load_xml,
    "load_json": load_json,
    # processors/transformers
    "flatten": flatten,
    "unflatten": unflatten,
    # filters
    "flake": flake,  # flatten key filter
    "key_filter": key_filter,  # filter dictionary key using glob pattern
    "xpath": xpath,  # XML xpath
    "match": match,  # similar to include
    "xml_flake": xml_flake,  # XML flatten key filter
    "lod_filter": lod_filter,  # list of dictionaries filter
    "jmespath": jmespath,  # filter structured data using jmespath lib
    "find": find,
    # transformers
    "xml_to_json": xml_to_json,
    "xml_flatten": xml_flatten,
    "xml_rm_ns": xml_rm_ns,
    "path": path_,
    "iplkp": iplkp,
}

task_instance_completed_dispatcher_multiresult = {
    # functions in this dictionary get MultiResult object and kwargs
    #
    # parsers
    "run_ttp": run_ttp,
}

task_instance_completed_dispatcher_per_result = {
    # functions in this dictionary get Result, Task, Host objects and kwargs
    #
    # parsers
    "ntfsm": ntfsm,
    "parse_ttp": parse_ttp,
}

task_started_dispatcher = {
    "run_ttp": add_commands_from_ttp_template,
}


class DataProcessor:
    """
    DataProcessor can process data obtained from devices. It is capable of:

    * loading data to Python structure from json, yaml, xml, csv or python format
    * serializing structured data to text in json, yaml, xml, csv or python format
    * filtering structured or text data
    * flattening and un-flattening nested data
    * parsing textual data into structured format

    :param dp: (list) list of Data Processors function names to pass results through

    ``dp`` argument can be of one of these types:

    * comma separated string of function names
    * list of function name strings and/or dictionaries with function details

    Data processor function dictionary items should have this structure::

        {
            "fun": function name [str],
            "k1": "v1", ... "kN": "vN"
        }

    Where:

    * ``fun`` - Reference Name of DataProcessor function to run
    * ``kN`` - Any additional key-word arguments to use with function
    """

    def __init__(self, dp=[]):
        self.dp = []
        if isinstance(dp, list):
            for i in dp:
                if isinstance(i, str):
                    self.dp.append({"fun": i.strip()})
                elif isinstance(i, dict):
                    self.dp.append(i)
                else:
                    raise TypeError(
                        "nornir_salt:DataProcessor dp list items should be dictionary or string not '{}'".format(
                            type(i)
                        )
                    )
        elif isinstance(dp, str):
            self.dp = [{"fun": i.strip()} for i in dp.split(",")]
        else:
            raise TypeError(
                "nornir_salt:DataProcessor dp argument should be list or string not '{}'".format(
                    type(dp)
                )
            )

    def task_started(self, task: Task) -> None:
        """ Pre-Process Task details before executing it """
        for dp_dict in self.dp:
            dp_dict_copy = dp_dict.copy()
            try:
                fun = dp_dict_copy.pop("fun")
                if fun in task_started_dispatcher:
                    task_started_dispatcher[fun](task, **dp_dict_copy)
            except:
                log.error(
                    "nornir-salt:DataProcessor task pre-processing task {} dp '{}' error:\n{}".format(
                        task, dp_dict, traceback.format_exc()
                    )
                )

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        """ Process results from device once main task completed """
        # check if has failed tasks, do nothing in such a case
        if result.failed:
            log.error("nornir_salt:DataProcessor do nothing, return, has failed tasks")
            return

        # run DataProcessor function
        for dp_dict in self.dp:
            dp_dict_copy = dp_dict.copy()
            try:
                fun = dp_dict_copy.pop("fun")
                if fun in task_instance_completed_dispatcher_multiresult:
                    try:
                        task_instance_completed_dispatcher_multiresult[fun](
                            result, task=task, **dp_dict_copy
                        )
                    except:
                        log.exception(
                            "nornir-salt:DataProcessor host '{}' function '{}' per-multiresult error".format(
                                host.name, fun
                            )
                        )
                elif fun in task_instance_completed_dispatcher_per_result_data:
                    for i in result:
                        try:
                            # check if need to skip this task
                            if hasattr(i, "skip_results") and i.skip_results is True:
                                continue
                            # pass task result through dp function
                            i.result = (
                                task_instance_completed_dispatcher_per_result_data[fun](
                                    i.result, **dp_dict_copy
                                )
                            )
                        except:
                            i.exception = traceback.format_exc()
                            log.exception(
                                "nornir-salt:DataProcessor host '{}' function '{}' per-task error".format(
                                    host.name, fun
                                )
                            )
                elif fun in task_instance_completed_dispatcher_per_result:
                    for i in result:
                        try:
                            # check if need to skip this task
                            if hasattr(i, "skip_results") and i.skip_results is True:
                                continue
                            # pass task result through dp function
                            task_instance_completed_dispatcher_per_result[fun](
                                result=i, task=task, host=host, **dp_dict_copy
                            )
                        except:
                            i.exception = traceback.format_exc()
                            log.exception(
                                "nornir-salt:DataProcessor host '{}' function '{}' per-result error".format(
                                    host.name, fun
                                )
                            )
                else:
                    raise KeyError(fun)
            except:
                log.exception(
                    "nornir-salt:DataProcessor host '{}' dp '{}'".format(
                        host.name, dp_dict
                    )
                )

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore subtasks

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass  # ignore subtasks

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        pass  # ignore final results



from .tcp_ping import tcp_ping
from .netmiko_send_commands import netmiko_send_commands
from .netmiko_send_command_ps import netmiko_send_command_ps
from .nr_test import nr_test
from .ncclient_call import ncclient_call
from .scrapli_netconf_call import scrapli_netconf_call
from .salt_cfg_gen import salt_cfg_gen
from .scrapli_send_commands import scrapli_send_commands
from .netmiko_send_config import netmiko_send_config
from .napalm_configure import napalm_configure
from .scrapli_send_config import scrapli_send_config
from .http_call import http_call
from .files import files, file_read, file_list, file_remove, file_diff
from .connections import connections, conn_close, conn_list
from .pygnmi_call import pygnmi_call
from .salt_clear_hcache import salt_clear_hcache
from .napalm_send_commands import napalm_send_commands
from .sleep import sleep

__all__ = (
    "tcp_ping",
    "netmiko_send_commands",
    "netmiko_send_command_ps",
    "nr_test",
    "ncclient_call",
    "scrapli_netconf_call",
    "salt_cfg_gen",
    "scrapli_send_commands",
    "netmiko_send_config",
    "napalm_configure",
    "scrapli_send_config",
    "http_call",
    "file_read",
    "file_list",
    "file_remove",
    "file_diff",
    "files",
    "connections",
    "conn_close",
    "conn_list",
    "pygnmi_call",
    "salt_clear_hcache",
    "napalm_send_commands",
    "sleep",
)



"""
tcp_ping
########

Tests connection to a TCP port trying to establish a three way
handshake. Useful for network discovery or testing.
    
tcp_ping sample usage
=====================

Sample code to run ``tcp_ping`` task::

    import pprint
    from nornir import InitNornir
    from nornir_salt.plugins.tasks import tcp_ping
    from nornir_salt.plugins.functions import ResultSerializer
    
    nr = InitNornir(config_file="config.yaml")
    
    result = NornirObj.run(
        task=tcp_ping,
        ports=[22]
    )

    result_dictionary = ResultSerializer(result)
    
    pprint.pprint(result_dictionary)
    
    # prints:
    # 
    # {'IOL1': {'tcp_ping': {22: True}},
    #  'IOL2': {'tcp_ping': {22: True}}}
    

tcp_ping returns
================

Returns dictionary of port numbers as keys with True/False as values

tcp_ping reference
==================

.. autofunction:: nornir_salt.plugins.tasks.tcp_ping.tcp_ping
"""

import socket
from typing import Optional, List

from nornir.core.task import Result, Task


def tcp_ping(
    task: Task, ports: List[int] = [], timeout: int = 1, host: Optional[str] = None
) -> Result:
    """
    :param ports: list of int, optional, tcp ports to ping, defaults to host's port or 22
    :param timeout: int, optional, connection timeout, defaults to 1
    :param host: string, optional, address to TCP ping, defaults to hosts's ``hostname`` value
    :returns: dictionary of port numbers as keys with True/False as values
    """
    if not ports:
        ports = [task.host.port if task.host.port else 22]
    elif isinstance(ports, int):
        ports = [ports]

    if isinstance(ports, list):
        if not all(isinstance(port, int) for port in ports):
            raise ValueError("Invalid value for 'ports'")
    else:
        raise ValueError("Invalid value for 'ports'")

    host = host or task.host.hostname

    result = {}
    for port in ports:
        s = socket.socket()
        s.settimeout(timeout)
        try:
            status = s.connect_ex((host, port))
            if status == 0:
                connection = True
            else:
                connection = False
        except (socket.gaierror, socket.timeout, socket.error):
            connection = False
        finally:
            s.close()
        result[port] = connection

    return Result(host=task.host, result=result)



"""
ToFileProcessor Plugin
######################

Processor plugin to save task execution results to file.

ToFileProcessor Sample Usage
============================

Code to demonstrate how to use ``ToFileProcessor`` plugin::

    from nornir import InitNornir
    from nornir_netmiko import netmiko_send_command
    from nornir_salt import ToFileProcessor

    nr = InitNornir(config_file="config.yaml")

    nr_with_processor = nr.with_processors([
        ToFileProcessor(tf="config", base_url="./Output/")
    ])

    nr_with_processor.run(
        task=netmiko_send_command,
        command_string="show run"
    )

ToFileProcessor reference
=========================

.. autofunction:: nornir_salt.plugins.processors.ToFileProcessor.ToFileProcessor
"""
import logging
import time
import os
import json
import traceback
import random

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task

log = logging.getLogger(__name__)


class ToFileProcessor:
    """
    ToFileProcessor can save task execution results to file on a per host basis.
    If multiple tasks present, results of all of them saved in same file.

    :param tf: (str) name of the file groups content
    :param base_url: (str) OS path to folder where to save files, default "/var/nornir-salt/"
    :param max_files: (int) default is 5, maximum number of file for given ``tf`` file group
    :param index: (str) index filename to read and store files data into

    Files saved under ``base_url`` location, where individual filename formed using string::

        {tf}__{timestamp}__{rand}__{hostname}.txt

    Where:

    * tf - value of ``tf`` attribute
    * timestamp - ``%d_%B_%Y_%H_%M_%S`` time formatted string, e.g. "12_June_2021_21_48_11"
    * rand - random integer in range from 10 to 1000
    * hostname - ``name`` attribute of host

    In addition, ``tf_index_{index}.json`` file created under ``base_url`` to track files created
    using dictionary structure::

        {
            "config": {
                "IOL1": [
                    {
                        "filename": "./tofile_outputs/config__22_August_2021_14_08_33__IOL1.txt",
                        "tasks": {
                            "show run | inc ntp": {
                                "span": [0, 48],
                                "content_type": "str"
                        }
                    }
                ]
            }
        }

    Where ``config`` is ``tf`` attribute value, ``"show run | inc ntp": [0, 48]`` -
    ``show run | inc ntp`` task results span indexes inside
    ``./tofile_outputs/config__22_August_2021_14_08_33__IOL1.txt`` text file.

    ``tf_index_{index}.json`` used by other plugins to retrieve previous results for the task,
    it could be considered as a simplified index database.
    """

    def __init__(self, tf, base_url="/var/nornir-salt/", max_files=5, index=None):
        self.tf = tf
        self.base_url = base_url
        self.max_files = max(1, max_files)
        self.index = index or "common"

        self.aliases_file = os.path.join(
            base_url, "tf_index_{}.json".format(self.index)
        )
        self.aliases_data = {}  # dictionary to store tf_index_{index}.json content

        self._load_aliases()

    def _load_aliases(self):
        # create aliases file if does not exist
        if not os.path.exists(self.aliases_file):
            os.makedirs(os.path.dirname(self.aliases_file), exist_ok=True)
            with open(self.aliases_file, mode="w", encoding="utf-8") as f:
                f.write(json.dumps({}))

        # load aliases data
        with open(self.aliases_file, mode="r", encoding="utf-8") as f:
            self.aliases_data = json.loads(f.read())

    def _dump_aliases(self):
        # save aliases data
        with open(self.aliases_file, mode="w", encoding="utf-8") as f:
            f.write(
                json.dumps(
                    self.aliases_data, sort_keys=True, indent=4, separators=(",", ": ")
                )
            )

    def task_started(self, task: Task) -> None:
        pass  # ignore

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        """save to file on a per-host basis"""

        host_filename = "{tf}__{timestamp}__{rand}__{hostname}.txt".format(
            timestamp=time.strftime("%d_%B_%Y_%H_%M_%S"),
            rand=random.randint(0, 1000),
            hostname=host.name,
            tf=self.tf,
        )
        host_filename = os.path.join(self.base_url, host_filename)

        # add aliases data
        self.aliases_data.setdefault(self.tf, {})
        self.aliases_data[self.tf].setdefault(host.name, [])

        # save data to file and populate alias details for tasks
        os.makedirs(os.path.dirname(host_filename), exist_ok=True)
        with open(host_filename, mode="w", encoding="utf-8") as f:
            self.aliases_data[self.tf][host.name].insert(
                0,
                {
                    "filename": host_filename,
                    "tasks": {},
                    "timestamp": time.strftime("%d %b %Y %H:%M:%S %Z"),
                },
            )
            span_start = 0

            for i in result:
                # check if need to skip this task results
                exception = (
                    str(i.exception)
                    if i.exception is not None
                    else i.host.get("exception", None)
                )
                if (
                    hasattr(i, "skip_results")
                    and i.skip_results is True
                    and not exception
                ):
                    continue
                # save results to file
                if isinstance(i.result, (str, int, float, bool)):
                    result_to_save = str(i.result)
                    self.aliases_data[self.tf][host.name][0]["tasks"][i.name] = {
                        "content_type": "str"
                    }
                # convert structured data to json
                else:
                    result_to_save = json.dumps(
                        i.result, sort_keys=True, indent=4, separators=(",", ": ")
                    )
                    self.aliases_data[self.tf][host.name][0]["tasks"][i.name] = {
                        "content_type": "json"
                    }
                f.write(result_to_save + "\n")

                # add aliases data
                span = (span_start, span_start + len(result_to_save) + 1)
                self.aliases_data[self.tf][host.name][0]["tasks"][i.name]["span"] = span
                span_start += len(result_to_save) + 1  # f.write appends \n hence +1

        # check if need to delete old files
        if len(self.aliases_data[self.tf][host.name]) > self.max_files:
            file_to_rm = self.aliases_data[self.tf][host.name].pop(-1)
            try:
                os.remove(file_to_rm["filename"])
            except:
                log.error(
                    "nornir-salt:ToFileProcessor failed to remove file '{}':\n{}".format(
                        file_to_rm, traceback.format_exc()
                    )
                )

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass  # ignore subtasks

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass  # ignore subtasks

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        self._dump_aliases()
